/*!
    \file   df_log.h
    \author Bernd Ottmann
    \date   21.02.2003
    \brief  Deklarationen der Funktionen für das Logging der Bibliothek.
*/
#ifndef DF_LOG_H
#define DF_LOG_H

#include "df_defines.h"

/*! 
    \brief Schaltet die Logfunktionalitaet der DLL ein.

    Loggen von Kommunikationsablaeufen in die angegebene Datei.

    \param filePath Zeiger auf die Zeichenfolge, welche den Pfad und/oder Dateinamen enthaelt.

    \return Ergebnis der Dateioeffnung.
    \retval 0 Datei konnte nicht erstellt oder geoeffnet werden.
    \retval 1 Logfunktionalitaet wurde erfolgreich eingeschaltet.

    \deprecated Ab der DLL - Version 2.0.3 wird diese Funktion nicht mehr unterstuetzt.
    \warning Verwenden Sie bitte statt dessen die Funktionen DFCSetLogFileName() / DFCSetLogOn().
*/
DF_API_EXTERN int DF_API_CALL DFCLogOn( char *filePath );

#if (DF_API_VERSION >= 0x02000300)


/*! 
    \brief Schaltet die Logfunktionalitaet des Kanals ein.

    Loggen von Kommunikationsablaeufen in die angegebene Datei.

    \param connectionId Kommunikationskanalnummer.

    \return Keine.

    \version Ab DLL - Version 2.0.3 enthalten.
*/
DF_API_EXTERN void DF_API_CALL DFCSetLogOn( int connectionId );

#endif


/*! 
    \brief Schaltet die Logfunktionalitaet der DLL aus.

    Abschalten des Logvorganges.

    \return Keine.

    \warning Ab der DLL - Version 2.0.3 wird diese Funktion nicht mehr unterstuetzt.
    \warning Verwenden Sie bitte statt dessen die Funktion DFCSetLogOff().
*/
DF_API_EXTERN void DF_API_CALL DFCLogOff();

#if (DF_API_VERSION >= 0x02000300)


/*! 
    \brief Schaltet die Logfunktionalitaet des Kanals aus.

    Abschalten des Logvorganges.

    \param connectionId Kommunikationskanalnummer.

    \return Keine.

    \version Ab DLL - Version 2.0.3 enthalten.
*/
DF_API_EXTERN void DF_API_CALL DFCSetLogOff( int connectionId );

#if (DF_API_VERSION >= 0x02000300)
/*! 
    \brief Setzen des Dateinamen fuer die Logfunktionalitaet des Kanals.

    Bekanntgeben der Datei die fuer das Loggen verwendet wird.

    \param connectionId Kommunikationskanalnummer.
    \param filePath Zeiger auf die Zeichenfolge, welche den Pfad und/oder Dateinamen enthaelt.

    \return Keine.

    \version Ab DLL - Version 2.0.3 enthalten.
*/
DF_API_EXTERN void DF_API_CALL DFCSetLogFileName( int connectionId, const unsigned char *filePath );
#endif


#endif

#endif
