/*!
    \file   df_lua_setup_records.h
    \author Bernd Ottmann
    \date   21.02.2003
    \brief  Deklarationen für die Funktionen der Setupdatensätze.
*/
#ifndef DF_LUA_SETUP_RECORDS_H
#define DF_LUA_SETUP_RECORDS_H

#include "df_lua_api.h"

#if (DF_API_VERSION >= 0x04030600)
/*! 
    \brief Datenzeiger ruecksetzen.

    Die Datenzeiger im Geraet werden rueckgesetzt. Die im Geraet eventuell enthaltenen  
    Datensaetze werden dadurch geloescht. 

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden.
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 2.0.0 enthalten.
*/
//DF_API_EXTERN int DF_API_CALL DFCComClearData( int connectionId, int deviceAddress );
int l_resetRecords( lua_State *L );

/*! 
    \brief Tabellenbeschreibungen fuer Datensaetze aus Geraet lesen.

    Es wird versucht die Tabellenbeschreibungen aus dem Geraet zu lesen. Hiernach koennen   
    mit den Funktionen DFCDatBCnt(), DFCDatBDatensatz() und DFCDatBFeld() auf den 
    ermittelten Daten gearbeitet werden.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden.
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 1.1.i enthalten.
*/
//DF_API_EXTERN int DF_API_CALL DFCLoadDatensatzbeschreibung( int connectionId, int deviceAddress, int *errorNumber );
int l_readRecordTables( lua_State *L );

/*! 
    \brief Anzahl der Tabellenbeschreibungen fuer Datensaetze ermitteln.

    Auf Grundlage der zuvor mit DFCLoadDatensatzbeschreibung() gelesenen Daten, wird 
    die Anzahl der definierten Tabellen zurueckgegeben.

    \param connectionId Kommunikationskanalnummer.

    \return Anzahl der mit DFCLoadDatensatzbeschreibung() gelesenen Tabellendefinitionen.

    \version Ab DLL - Version 1.1.i enthalten.
*/
//DF_API_EXTERN int DF_API_CALL DFCDatBCnt( int connectionId );
int l_getRecordTableCount( lua_State *L );

/*! 
    \brief Grunddaten einer Tabellenbeschreibungen fuer Datensaetze ermitteln.

    Auf Grundlage der zuvor mit DFCLoadDatensatzbeschreibung() gelesenen Daten, werden 
    die Grunddaten zu einer Tabellen zurueckgegeben.

    \param connectionId Kommunikationskanalnummer.
    \param number Tabellennummer im Bereich von (0 - (DFCDatBCnt() - 1)).
    \param name Zeiger auf Speicherbereich fuer Tabellenname. (Mindestens 17 Zeichen).
    \param fieldCount Zeiger auf Variable, welche die Feldanzahl der Tabelle erhaelt.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden. (Tabelle nicht definiert!)
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 1.1.i enthalten.
*/
//DF_API_EXTERN int DF_API_CALL DFCDatBDatensatz( int connectionId, int number, unsigned char *name, int *fieldCount );
int l_getRecordTable( lua_State *L );

/*! 
    \brief Grunddaten eines Tabellenfeldes fuer Datensaetze ermitteln.

    Auf Grundlage der zuvor mit DFCLoadDatensatzbeschreibung() gelesenen Daten, werden 
    die Grunddaten zu einem Tabellenfeld zurueckgegeben.

    \param connectionId Kommunikationskanalnummer.
    \param datBNumber Tabellennummer im Bereich von (0 - (DFCDatBCnt() - 1)).
    \param number Feldnummer.
    \param name Zeiger auf Speicherbereich fuer Feldname. (Mindestens 17 Zeichen).
    \param type Zeiger auf Variable, welche den Feldtyp erhaelt.
    \param size Zeiger auf Variable, welche die definierte Laenge des Feldes erhaelt,
                    in dieser ist auch das Nullterminierungszeichen enthalten.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden. (Tabelle nicht definiert!)
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 1.1.i enthalten.
*/
//DF_API_EXTERN int DF_API_CALL DFCDatBFeld( int connectionId, int datBNumber, int number, unsigned char *name, int *type, int *size );
int l_getRecordField( lua_State *L );

//DF_API_EXTERN int DF_API_CALL DFCReadRecord( int channelID, int deviceAddress, char *buffer, int *bufferSize, int *errorNumber );
int l_readRecord( lua_State *L );

//DF_API_EXTERN int DF_API_CALL DFCQuitRecord( int channelID, int deviceAddress, int *errorNumber );
int l_quitRecord( lua_State *L );

//DF_API_EXTERN int DF_API_CALL DFCRestoreRecords( int connectionId, int deviceAddress, int *errorNumber );
int l_restoreRecords( lua_State *L );

#endif

#endif
