#include "../df_lua_active_mode.h"
#include <algorithm>

//DF_API_EXTERN int DF_API_CALL DFCStartActiveConnection( char *networkAddress, int listenPort, int deviceAddress, int timeout, int aliveTimeout, int maxConnections, int infoFlags, int unused_setTo0, int *errorNumber )
int l_startActiveMode( lua_State *L )
{
    // networkAddress
    const char *networkAddress = luaL_checkstring( L, 1 );
    // listenPort
    int listenPort = static_cast<int>( luaL_checkinteger( L, 2 ) );
    // timeout
    int timeout = static_cast<int>( luaL_checkinteger( L, 3 ) );
    // aliveTimeout
    int aliveTimeout = static_cast<int>( luaL_checkinteger( L, 4 ) );
    // maxConnections
    int maxConnections = static_cast<int>( luaL_checkinteger( L, 5 ) );
    // infoFlags
    int infoFlags = static_cast<int>( luaL_checkinteger( L, 6 ) );
    // unused_setTo0
    int unused_setTo0 = static_cast<int>( luaL_optinteger( L, 7, 0 ) );

    int res = 1;
    int errorNumber;
    lua_pushboolean( L, DFCStartActiveConnection( const_cast<char *>( networkAddress ), listenPort, 0, timeout, aliveTimeout, maxConnections, infoFlags, unused_setTo0, &errorNumber ) );
    if ( lua_toboolean( L, -1 ) == false )
        push_globalErrorNumber( errorNumber, res );

    return res;
}

//DF_API_EXTERN int DF_API_CALL DFCStopActiveConnection( int *errorNumber )
int l_stopActiveMode( lua_State *L )
{
    int res = 1;
    int errorNumber;
    lua_pushboolean( L, DFCStopActiveConnection( &errorNumber ) );
    if ( lua_toboolean( L, -1 ) == false )
        push_globalErrorNumber( errorNumber, res );

    return res;
}

//DF_API_EXTERN int DF_API_CALL DFCGetFirstActiveConnectionId()
int l_getActiveModeFirstId( lua_State *L )
{
    lua_pushinteger( L, DFCGetFirstActiveConnectionId() );

    return 1;
}

//DF_API_EXTERN int DF_API_CALL DFCGetNextActiveConnectionId( int previousId )
int l_getActiveModeNextId( lua_State *L )
{
    int previousId = static_cast<int>( luaL_checkinteger( L, 1 ) );
    luaL_argcheck( L, previousId >= 1 && previousId < (MAX_CHANNELID - 1), 1, "out of range" );

    lua_pushinteger( L, DFCGetNextActiveConnectionId( previousId ) );

    return 1;
}

//DF_API_EXTERN int DF_API_CALL DFCGetInfoActiveChannel( int connectionId, char *value, int *valueLength )
int l_getActiveModeInfo( lua_State *L )
{
    int channelId = static_cast<int>( luaL_checkinteger( L, 1 ) );
    luaL_argcheck( L, channelId >= 1 && channelId < (MAX_CHANNELID - 1), 1, "out of range" );

    std::string info( 1024, '\0' );
    int infoLength = info.size();
    int res = DFCGetInfoActiveChannel( channelId, const_cast<char *>( info.data() ), &infoLength );
    if ( res == 0 )
    {
        lua_pushnil( L );
        res = 1;
    }
    else
    {
        std::replace( info.begin(), info.end(), '\r', '\t' );
        lua_pushlstring( L, info.data(), info.find_first_of( '\0' ) );
        res = 1;
    }

    return res;
}

//DF_API_EXTERN int DF_API_CALL DFCSetRecordAvailable( int enabled )
int l_setActiveModeNotify( lua_State *L )
{
    int enabled = static_cast<int>( luaL_checkinteger( L, 1 ) );

    lua_pushinteger( L, DFCSetRecordAvailable( enabled ) );

    return 1;
}

//DF_API_EXTERN int DF_API_CALL DFCRecordAvailable( int *connectionId, int *deviceAddress, char *infoText, int *infoTextLength, int *errorNumber )
int l_getActiveModeNotifyAvailable( lua_State *L )
{
    int channelId;
    int deviceAddress;
    std::string info( 1024, '\0' );
    int infoLength = info.size();
    int res = 1;
    int errorNumber;
    lua_pushboolean( L, DFCRecordAvailable( &channelId, &deviceAddress, const_cast<char *>( info.data() ), &infoLength, &errorNumber ) );
    if ( lua_toboolean( L, -1 ) == false )
    {
        lua_pushnil( L );
        res++;
        lua_pushnil( L );
        res++;
        lua_pushnil( L );
        res++;
        push_globalErrorNumber( errorNumber, res );
    }
    else
    {
        lua_pushinteger( L, channelId );
        res++;
        lua_pushinteger( L, deviceAddress );
        res++;
        std::replace( info.begin(), info.end(), '\r', '\t' );
        lua_pushlstring( L, info.data(), info.find_first_of( '\0' ) );
        res++;
    }

    return res;
}

//DF_API_EXTERN int DF_API_CALL DFCBindDeviceToChannel( int connectionId, int deviceType, int deviceSerial, char *deviceIp, int *errorNumber )
int l_bindActiveModeChannel( lua_State *L )
{
    int channelId =  static_cast<int>( luaL_checkinteger( L, 1 ) );
    int deviceType =  static_cast<int>( luaL_checkinteger( L, 2 ) );
    int deviceSerial =  static_cast<int>( luaL_checkinteger( L, 3 ) );
    const char *deviceIp =  luaL_checkstring( L, 4 );

    int res = 1;
    int errorNumber;
    lua_pushboolean( L, DFCBindDeviceToChannel( channelId, deviceType, deviceSerial, const_cast<char *>( deviceIp ), &errorNumber ) );

    if ( lua_toboolean( L, -1 ) == false )
        push_globalErrorNumber( errorNumber, res );

    return res;
}