/**
 * @file  Channel1_2".cpp
 * @brief Implementierung der Klasse CChannel1_2.
 */
//////////////////////////////////////////////////////////////////////

#include "Channel1_2.h"

// Versionspruefung
#include "../Common/DFComUtils.h"
#include "../../DatafoxLibraryIV/libraryversion.h"
#include "../Common/eeprom.h"
#include <sys/types.h>
#include <sys/stat.h>
#include "../Common/softwareinfo.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

void CChannel1_2::InitClass(short channelID, short deviceID)
{
    (void)channelID;
    // Kommunikationsklasse
    m_channelBase = NULL;

    // Protokollversion Unbestimmt
    m_protocolVersion = 0;

    // Busnummer
    m_deviceID = (unsigned char)deviceID;
    m_deviceIDRemote = 0;

    // CallBack nicht definiert
    m_callBack1 = NULL;
}

CChannel1_2::~CChannel1_2()
{

}

//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

unsigned short CChannel1_2::BinToUShort(unsigned char *Buf)
{
    unsigned short Val;

    Val = Buf[0];
    Val <<= 8;
    Val += Buf[1];
    return Val;
}

unsigned int CChannel1_2::BinToULong(unsigned char *Buf)
{
    unsigned int Val;

    Val = Buf[0];
    Val <<= 8;
    Val += Buf[1];
    Val <<= 8;
    Val += Buf[2];
    Val <<= 8;
    Val += Buf[3];
    return Val;
}

// Konvertierung MSB First
void CChannel1_2::UShortToBin(unsigned short Val,unsigned char *Buf)
{
    Buf[0] = (unsigned char)(Val >> 8);
    Buf[1] = (unsigned char)(Val & 0xFF);
}

// Konvertierung MSB First
void CChannel1_2::ULongToBin(unsigned int Val,unsigned char *Buf)
{
    Buf[0] = (unsigned char)((Val >> 24)&0xFF);
    Buf[1] = (unsigned char)((Val >> 16)&0xFF);
    Buf[2] = (unsigned char)((Val >> 8)&0xFF);
    Buf[3] = (unsigned char)(Val & 0xFF);
}

// Sende und Empfang von Modbus - Paket
int CChannel1_2::Command(unsigned char *pucTxBuf, int iTxCount, 
                        unsigned char *pucRxBuf, int *piRxCount, 
                        int *piError)
{
    // Bei 2er Protokollversion verzweigen.
    if (m_protocolVersion == 2) {

        int i;
        unsigned char buf[MAX_MB_PAKET];

        for (i=0; i < iTxCount; i++) {

            buf[i] = pucTxBuf[i];
        }

        if (!NewMB(GetDeviceIDRemote(), buf, &i, piError)) {

            return FALSE;
        }

        for (*piRxCount=0; *piRxCount < i; (*piRxCount)++) {

            pucRxBuf[*piRxCount] = buf[*piRxCount];
        }

        return TRUE;
    }

    // Paketlaenge
    unsigned int uiPaketLength;
    // Anzahl geschriebener oder gelesender Zeichen
    unsigned int ulReadWriteCount = 0;

    // Nicht aktuelle Busnummer
    if (pucTxBuf[0] != GetDeviceIDRemote()) { 

        GetChannelBase().trace("Command: Gesetzte Busnummer ist nicht die des zu sendenen Paketes.\n");	

        *piError = IV_ERROR_DEVICE_ID_REMOTE;
        return FALSE;
    }

//@@: Callback mit Parameter andenken.
    if (!ExeCallBack()) {
    
        *piError = IV_ERROR_USER_CANCEL;
        return FALSE;
    }


    // Befehl losschicken
    if (GetChannelBase().writeData((void*)pucTxBuf, iTxCount, (int&)ulReadWriteCount) != IV_ERROR_SUCCESS){
        
        GetChannelBase().trace("Command: Paket konnte nicht versendet werden.\n");	
        
        *piError = IV_ERROR_TRANSMIT;
        return FALSE;
    }

    if (ulReadWriteCount != (unsigned)iTxCount) {
    
        GetChannelBase().trace("Command: Paket wurde nicht vollstaendig versendet.(%i) von (%i)\n", ulReadWriteCount, iTxCount);	

        // zu wendig gesendet	
        *piError = IV_ERROR_TRANSMIT;
        return FALSE;
    }

    // Die ersten 4 Bytes der Antwort lesen
    if (GetChannelBase().readData((void*)pucRxBuf, 4, (int&)ulReadWriteCount, 4) != IV_ERROR_SUCCESS) {
            
        GetChannelBase().trace("Command: Warunung beim Lesen des Paketkopfes.(%i) von (%i)\n", ulReadWriteCount, 4);	
    }
    // Empfang pruefen
    if (ulReadWriteCount == 0) {
        
        GetChannelBase().trace("Command: Es wurde kein Paketkopf empfangen.\n");	
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }
    // Nachlesen?
    if (ulReadWriteCount < 4) {
    
        unsigned int lReadCount = 0;
        // Die ersten 4 Bytes der Antwort lesen
        if (GetChannelBase().readData((void*)&pucRxBuf[ulReadWriteCount], 4 - ulReadWriteCount, (int&)lReadCount, 2) != IV_ERROR_SUCCESS) {
                
            GetChannelBase().trace("Command: Das Nachlesen zum Paketkopf schlug fehl.(%i) von (%i)\n", lReadCount, (4 - ulReadWriteCount));	

            // AW1 kann nicht gelesen werden
            *piError = IV_ERROR_RECEIVE;
            return FALSE; 
        }

        ulReadWriteCount += lReadCount;
    }
    // Empfang pruefen
    if (ulReadWriteCount != 4) {
        
        GetChannelBase().trace("Command: Es wurden zu wenige Zeichen für den Paketkopf empfangen.(%i) von (%i)\n", ulReadWriteCount, 4);	

        // Schnittstelle fuer naechste Anfrage bereinigen
        GetChannelBase().flush();

        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    if (pucTxBuf[0] != pucRxBuf[0]) {

        GetChannelBase().trace("Command: Busnummer von Sender und Empfaenger ungleich Tx(%i): Rx(%i)\n", pucTxBuf[0], pucRxBuf[0]);	

        // Schnittstelle fuer naechste Anfrage bereinigen
        GetChannelBase().flush();

        // Absender falsch	
        *piError = IV_ERROR_ADDRESS;
        return FALSE; 
    }

    // Die Laenge vom Paket lesen
    uiPaketLength = BinToUShort(&pucRxBuf[1]);

    // Rest vom Paket lesen
    if(GetChannelBase().readData((void*)&pucRxBuf[4], uiPaketLength - 4, (int&)ulReadWriteCount, 5) != IV_ERROR_SUCCESS){

        GetChannelBase().trace("Command: Warnung beim Lesen der Paketdaten.(%i) von (%i)\n", ulReadWriteCount, uiPaketLength - 4);	
    }
    if (ulReadWriteCount == 0) {
    
        GetChannelBase().trace("Command: Es wurden keine Daten zum Paketlaenge empfangen.(%i) von (%i)\n", ulReadWriteCount, uiPaketLength - 4);	

        // AW2 zu kurz
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }
    // Nachlesen?
    if (ulReadWriteCount < (uiPaketLength - 4)) {
        
        unsigned int lReadCount = 0;
        // Die Paketdaten lesen
        if (GetChannelBase().readData((void*)&pucRxBuf[4 + ulReadWriteCount], (uiPaketLength - 4) - ulReadWriteCount, (int&)lReadCount, 2) != IV_ERROR_SUCCESS) {
                
            GetChannelBase().trace("Command: Das Nachlesen der Paketdaten schlug fehl.(%i) von (%i)\n", lReadCount, (uiPaketLength - 4) - ulReadWriteCount);	

            // AW1 kann nicht gelesen werden
            *piError = IV_ERROR_RECEIVE;
            return FALSE; 
        }

        ulReadWriteCount += lReadCount;
    }

    if (ulReadWriteCount != (uiPaketLength - 4)) {
        
        GetChannelBase().trace("Command: Die gelesene Paketlaenge stimmt nicht mit der angegebenen ueberein.(%i) von (%i)\n", ulReadWriteCount + 4, uiPaketLength);	

        // Schnittstelle fuer naechste Anfrage bereinigen
        GetChannelBase().flush();

        // AW2 zu kurz
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // Paketgroesse zurueckgeben
    *piRxCount = uiPaketLength;

    // Kommando erfolgreich ausgefuehrt
    return TRUE;

}

int CChannel1_2::Command2(unsigned char *sbuf,int cnt,unsigned char *ebuf,int &ret, int *Err)
{
    unsigned int nb = 0;
    int Cnt;

    if (!ExeCallBack()) {*Err = IV_ERROR_BREAK; return FALSE;}
    if (cnt) {
        if (GetChannelBase().writeData(sbuf,cnt,(int&)nb) != IV_ERROR_SUCCESS) {*Err = IV_ERROR_TRANSMIT; return FALSE;}
        if (nb != (unsigned)cnt) {*Err = IV_ERROR_TRANSMIT; return FALSE;}
    }

    if (ret > 0) {
        
        // Länge der Antwort ist bekannt
        if (GetChannelBase().readData(ebuf,ret,(int&)nb, 3) != IV_ERROR_SUCCESS) {*Err = IV_ERROR_RECEIVE; return FALSE;}
        if (nb == 0) {*Err = IV_ERROR_UNREACHABLE; return FALSE;}
        if (nb != (unsigned int)ret) {*Err = IV_ERROR_RECEIVE; return FALSE;}
        *Err = IV_ERROR_SUCCESS;
        GetChannelBase().trace("OK..\n");

    } else {

        // Laenge der Antwort ist nicht bekannt und muß aus dem Protokoll ermittelt werden
        //	Protokoll-Kopf lesen
        Cnt = 8;
        if (GetChannelBase().readData(ebuf, Cnt, (int&)nb, 4) != IV_ERROR_SUCCESS) {

            GetChannelBase().trace("WARNING_00: ReceiveData(%i != %i)\n", Cnt, nb);
        }
        if (nb == 0) {

            *Err = IV_ERROR_UNREACHABLE; 
            GetChannelBase().trace("ERROR_01: ReceiveData(%i != %i)\n", Cnt, nb);
            return FALSE;
        }
        // Nachlesen?
        if (nb > 0 && nb < (unsigned)Cnt) {

            unsigned int lnb = 0;
            // Noch nicht alle Byte gelesen -> versuchen weitere Bytes zu lesen
            if (GetChannelBase().readData(&ebuf[nb], Cnt-nb, (int&)lnb, 2) != IV_ERROR_SUCCESS) {
                
                *Err = IV_ERROR_RECEIVE; 
                GetChannelBase().trace("ERROR_02: ReceiveData.2 (%i != %i)\n", Cnt-nb, lnb);
                return FALSE;
            }

            // Nachgelesene Anzahl Zeichen
            nb += lnb;
        }
        if (nb != (unsigned int)Cnt) {

            *Err = IV_ERROR_RECEIVE; 
            GetChannelBase().trace("ERROR_03: ReceiveData(%i != %i)\n", Cnt, nb);
            return FALSE;
        }
    
        // Gesamtpaketlaenge
        Cnt = BinToUShort(&ebuf[V_O_LAENGE]);

        // Laenge auf plausibilitaet pruefen
        if (Cnt < 15 || Cnt > MAX_MB_PAKET_2) {

            *Err = IV_ERROR_PACKETLENGTH; 
            GetChannelBase().trace("ERROR_04: %i\n", Cnt);
            GetChannelBase().flush();
            return FALSE;
        }

        // Die restlichen Bytes lesen
        if (GetChannelBase().readData(&ebuf[8], Cnt-8, (int&)nb, 5) != IV_ERROR_SUCCESS) {

            GetChannelBase().trace("WARNING_01: ReceiveData(%i != %i)\n", Cnt-8, nb);
        }
        if (nb == 0) {

            *Err = IV_ERROR_UNREACHABLE; 
            GetChannelBase().trace("ERROR_01: ReceiveData(%i != %i)\n", Cnt, nb);
            return FALSE;
        }
        // Nachlesen?
        if (nb > 0 && nb < (unsigned)(Cnt-8)) {

            unsigned int lnb = 0;
            // Noch nicht alle Byte gelesen -> versuchen weitere Bytes zu lesen
            if (GetChannelBase().readData(&ebuf[nb], (Cnt-8)-nb, (int&)lnb, 2) != IV_ERROR_SUCCESS) {
                
                *Err = IV_ERROR_RECEIVE; 
                GetChannelBase().trace("ERROR_02: ReceiveData.2 (%i, %i)\n", Cnt-nb, lnb);
                return FALSE;
            }

            // Nachgelesene Anzahl Zeichen
            nb += lnb;
        }
        if (nb != (unsigned int)(Cnt-8)) {

            *Err = IV_ERROR_RECEIVE; 
            GetChannelBase().trace("ERROR_06: ReceiveData.2 (%i != %i)\n", Cnt-8, nb);
            return FALSE;
        }

        ret = Cnt;
        *Err = IV_ERROR_SUCCESS;
    }

    return TRUE;
}

// Einfache Anfrage ob Geraet mit Protokollversion 2 erreichbar.
int CChannel1_2::Request2(unsigned char deviceIDRemote)
{
    int Versuch,transmitCount,receiveCount;
    int len, *Len, err, *Err;
    Len = &len;
    Err = &err;

    // Laenge des alten Modbusprotokollpackets
    len = 0x0C;

    // Sendebuffer aufbereiten
    //	Adressen
    UShortToBin(deviceIDRemote, m_transmit);	// Adresse Empfänger
    UShortToBin(GetDeviceID(), &m_transmit[V_O_AD_ABSENDER_S2]);	// Adresse Sender

    // Paketlänge
    transmitCount = V_KOPF+1+1+1+2+1+len+2;	// Sendepaket
    receiveCount = -1;	// Empfangspaket
    UShortToBin(transmitCount,&m_transmit[V_O_LAENGE]);
    m_transmit[V_O_B_S2] = B_S2_NO;
    m_transmit[V_O_PAKET_NR] = 0;
    m_transmit[V_O_TIME_TO_LIVE] = 8;
    m_transmit[V_O_B_S3] = B_S3_NO;
    m_transmit[V_O_ANZ_BLOECKE] = 1;
    m_transmit[V_O_AKT_BLOCK] = 1;
    UShortToBin(GetDeviceID(), &m_transmit[V_O_AD_ABSENDER_S4]);	// Adresse Sender
    m_transmit[V_O_B_S4] = 0;//B_S4_LESEN|BB_S4_ANFRAGE;		// Anfrage lesen
    //ULongToBin(Add,&m_transmit[V_O_ADD]);

/*
    for (i=0;i<*Len;i++) {
        m_transmit[i+V_O_ADD] = buf[i];	// Daten in Sendebuffer kopieren
        // Bei OldMB entfällt die Daten-Adresse.
        // Daher sind die 4 für die Adresse reservierten Byte schon Daten
    }
*/
    // Altes Packet in neues, als Daten einbinden
    m_transmit[V_O_ADD + 0] = (unsigned char)deviceIDRemote;
    m_transmit[V_O_ADD + 1] = 0;
    m_transmit[V_O_ADD + 2] = 0x0C;		// Laenge des Paketes
    m_transmit[V_O_ADD + 3] = 0x21;//MB_READ;
    ULongToBin(0x42000000/*MB_GET_VERSION*/, &m_transmit[V_O_ADD + 4]);
    UShortToBin(2, &m_transmit[V_O_ADD + 8]);
    MkPaketCrc(&m_transmit[V_O_ADD], 10);

    // CRC - Gesamtpacket
    MkPaketCrc(m_transmit, transmitCount-2);

    // Daten lesen incl. Wiederholungen
    for (Versuch=A_VERSUCHE_OMB - 1;Versuch < A_VERSUCHE_OMB; Versuch++) {
        // Im Fehlerfall zur Sicherheit große Anzahl Zeichen lesen um Empfangsbuffer
        // des Betriebssystem zu lehren
        // Paket senden und Antwort abholen
        if (!Command2(m_transmit,transmitCount,m_receive,receiveCount,Err)) {
            continue;
        }
        // Antwort auswerten
        //	CRC prüfen
        if (!CheckPaketCrc(m_receive,receiveCount)) { 
            *Err = IV_ERROR_CRC_PACKET;
            continue;
        }
        //	Empfänger-/Absender-Adresse prüfen
        if (BinToUShort(m_receive) != GetDeviceID() || 
              BinToUShort(&m_receive[V_O_AD_ABSENDER_S2]) != deviceIDRemote ||
              BinToUShort(&m_receive[V_O_AD_ABSENDER_S4]) != deviceIDRemote) {
            *Err = IV_ERROR_ADDRESS;
            continue;
        }
        //	Längenfeld prüfen
        if (BinToUShort(&m_receive[V_O_LAENGE]) != receiveCount) {
            *Err = IV_ERROR_PACKETLENGTH;
            continue;
        }
        //	Befehl Schicht 4 prüfen
        if (m_receive[V_O_B_S4] != (B_S4_LESEN|BB_S4_ANTWORT)) {
            *Err = IV_ERROR_COMMAND;
            continue;
        }
        // Anzahl der gelesenen Datenbyte berechnen
        *Len = receiveCount - (V_KOPF+1+1+1+2+1+2);
        // Daten in Buffer der Applikation kopieren
//		for(i=0;i<*Len;i++) buf[i] = m_receive[i+V_O_ADD];
        *Err = IV_ERROR_SUCCESS;
    
        // Ermittelte Protokollversion halten
        m_protocolVersion = 2;

        return TRUE; // Daten erfolgreich gelesen
    }

    return FALSE;
}

int CChannel1_2::NewMB(unsigned short deviceIDRemote,unsigned char *buf, int *Len, int *Err)
// Mod-Bus Protokoll (AEIII+) in neues Protokoll verpacken, senden und Antwort zurückgeben
//  AdEmpfaenger -> Adresse (Gerätenummer) des Empfängers
//	buf	-> Buffer für Mod-Bus-Paket senden und empfangen
//	Len	-> Anzahl zu sende Byte. Nach dem Aufruf -> Anzahl Empfangene Byte
//	Err -> Bei einem Fehler -> Fehlernummer
//  return -> 1 -> senden/empfangen erfolgreich
//			  0 -> Fehler -> Fehlernummer in Var. Err	
{
    int Versuch,transmitCount,receiveCount,i;

    // Sendebuffer aufbereiten
    //	Adressen
    UShortToBin(deviceIDRemote, m_transmit);	// Adresse Empfänger
    UShortToBin(GetDeviceID(),&m_transmit[V_O_AD_ABSENDER_S2]);	// Adresse Sender
    // Paketlänge
    transmitCount = V_KOPF+1+1+1+2+1+*Len+2;	// Sendepaket
    receiveCount = -1;	// Empfangspaket
    UShortToBin(transmitCount,&m_transmit[V_O_LAENGE]);
    m_transmit[V_O_B_S2] = B_S2_NO;
    m_transmit[V_O_PAKET_NR] = 0;
    m_transmit[V_O_TIME_TO_LIVE] = 5;
    m_transmit[V_O_B_S3] = B_S3_NO;
    m_transmit[V_O_ANZ_BLOECKE] = 1;
    m_transmit[V_O_AKT_BLOCK] = 1;
    UShortToBin(GetDeviceID(),&m_transmit[V_O_AD_ABSENDER_S4]);	// Adresse Sender
    m_transmit[V_O_B_S4] = 0;//B_S4_LESEN|BB_S4_ANFRAGE;		// Anfrage lesen
    //ULongToBin(Add,&m_transmit[V_O_ADD]);
    for (i=0;i<*Len;i++) {
        m_transmit[i+V_O_ADD] = buf[i];	// Daten in Sendebuffer kopieren
        // Bei OldMB entfällt die Daten-Adresse.
        // Daher sind die 4 für die Adresse reservierten Byte schon Daten
    }
    MkPaketCrc(m_transmit,transmitCount-2);
    // Daten lesen incl. Wiederholungen
    for (Versuch=0;Versuch < A_VERSUCHE_OMB; Versuch++) {
        // Im Fehlerfall zur Sicherheit große Anzahl Zeichen lesen um Empfangsbuffer
        // des Betriebssystem zu lehren
        if (Versuch >= 2) Command2(m_transmit,0,m_receive,i=MAX_MB_PAKET_2,Err);
        // Paket senden und Antwort abholen
        if (!Command2(m_transmit,transmitCount,m_receive,receiveCount,Err)) {
            continue;
        }
        // Antwort auswerten
        //	CRC prüfen
        if (!CheckPaketCrc(m_receive,receiveCount)) { 
            *Err = IV_ERROR_CRC_PACKET;
            continue;
        }
        //	Empfänger-/Absender-Adresse prüfen
        if (BinToUShort(m_receive) != GetDeviceID() || 
              BinToUShort(&m_receive[V_O_AD_ABSENDER_S2]) != deviceIDRemote ||
              BinToUShort(&m_receive[V_O_AD_ABSENDER_S4]) != deviceIDRemote) {
            *Err = IV_ERROR_ADDRESS;
            continue;
        }
        //	Längenfeld prüfen
        if (BinToUShort(&m_receive[V_O_LAENGE]) != receiveCount) {
            *Err = IV_ERROR_PACKETLENGTH;
            continue;
        }
        //	Befehl Schicht 4 prüfen
        if (m_receive[V_O_B_S4] != (B_S4_LESEN|BB_S4_ANTWORT)) {
            *Err = IV_ERROR_COMMAND;
            continue;
        }
        // Anzahl der gelesenen Datenbyte berechnen
        *Len = receiveCount - (V_KOPF+1+1+1+2+1+2);
        // Daten in Buffer der Applikation kopieren
        for(i=0;i<*Len;i++) buf[i] = m_receive[i+V_O_ADD];
        *Err = IV_ERROR_SUCCESS;
        return TRUE; // Daten erfolgreich gelesen
    }
    return FALSE;
}

// Testen ob PZE - Master IV im Bootloader oder Applicationsmodus ist.
// return
// 0 -> Kommunikationsfehler
// 1 -> Applicationsmodus
// 2 -> Bootloadermodus
int CChannel1_2::GetModus(int *error)
{
    unsigned char tmp[2];

    // Befehl ist nur im Zusammenhang mit AEIV sinnvoll.
//	if (GetChannelType() != ClassSerial && GetChannelType() != ClassSocket) {

//		*error = IV_ERROR_CHANNEL_CLOSED;
//		return FALSE;
//	}

    // Allgemeine Erreichbarkeit pruefen
    if (ReadXByte(0, 2, tmp, error)) {

        // Der Befehl wird vom Bootloader nicht unterstuetzt
        // MB_GET_VERSION
        if (ReadXByte(0x42000000, 2, tmp, error)) {

            return 2;
        }

        return 1;
    }

    return FALSE;
}

// CRC berechnen und an das Ende des Buffers anhängen
void CChannel1_2::MkPaketCrc(unsigned char *buf,int cnt)	
{
    int i,j;
    unsigned short int crc;
  
    crc = 0xFFFF;
    for (i=0;i<cnt;i++) {
        crc = crc ^ buf[i];
        for (j=0;j<8;j++) {
            if (crc & 0x1) {crc >>= 1; crc = crc^0xA001;} else crc >>= 1;
        }
    }
    buf[cnt] = (unsigned char)(crc&0xFF);
    buf[cnt+1] = (unsigned char)(crc>>8);
}

int CChannel1_2::CheckPaketCrc(unsigned char *buf, int cnt)
{
    unsigned short int crc;
    int i,j;

    crc = 0xFFFF;
    for (i=0;i<cnt-2;i++) {
        crc = crc ^ buf[i];
        for (j=0;j<8;j++) {
            if (crc & 0x1) {crc >>= 1; crc = crc^0xA001;} else crc >>= 1;
        }
    }
    if (buf[cnt-2]!=(unsigned char)(crc&0xFF)) return FALSE;
    if (buf[cnt-1]!=(unsigned char)(crc>>8)) return FALSE;
    return TRUE;
}

// Schreiben in AEIII+ (High-Level)
// Add -> Byte Adresse
// Cnt -> Anzahl Zeichen (max. 1MByte)
// Buf -> Buffer fuer empfangene Zeichen
int CChannel1_2::WriteMem(unsigned int Add, int Cnt, unsigned char *Buf, int *piError, int MaxB)
{
    int i, j, ac, End = FALSE, ok;

    // Fehler loeschen
    *piError = IV_ERROR_SUCCESS;

    if (m_protocolVersion == 2) {

        if (MaxB == 0) {
            // Wenn EEPROM geschrieben wird
            MaxB = 8;
        } else if (MaxB == 32) {

        } else {
            MaxB = MAX_MB_PAKET_DATA;
        }
    } else {

        // Bis zu (Gesamtpaket[ 0xFF ] - Overhead[ MAX_MB_PAKET_HEADER ]) Byte Daten in Geraet schreiben
        if (MaxB > MAX_MB_PAKET_DATA) {
        
            *piError = IV_ERROR_FUNCTION_PARAMETER;
            return FALSE;
        }
    }

    // Daten in Pakete zerlegen
    for (i=0; !End; i += MaxB) {

        // Aufteilen
        if ((Cnt - i) < MaxB) {

            ac = Cnt - i;
        } else {

            ac = MaxB;
        }

        // Noch zu versendende Daten
        if ( ac > 0 ) {

            ok = FALSE;
            
            // Senden in mehreren Versuchen
            for (j=0; j < 4 && !ok; j++) {

                if (WriteXByte(Add + i, ac, &Buf[i], piError)) {

                    ok = TRUE;
                } else {

                    // Abbruch durch CallBack!
                    if (*piError == IV_ERROR_USER_CANCEL) {
                        
                        return FALSE;
                    }

                    Sleep(50);
                }
            }

            // Senden Fehlgeschlagen?
            if (!ok) {

                // Letzten Fehler zurueckgeben
                return FALSE;
            }

        } else {

            // Alle versendet
            End = TRUE;
        }
    }

    return TRUE;
}

// X Byte in AEIII+ schreiben (Low-Level)
// Add -> Byte Adresse
// Cnt -> Anzahl Zeichen (max. 242 Byte)
// Buf -> Buffer fuer empfangene Zeichen
int CChannel1_2::WriteXByte (unsigned int Add, int Cnt, unsigned char *Buf, int *piError)
{
    int ret, i, h;
    unsigned char WBuf[MAX_WBUF];
    unsigned char RBuf[MAX_RBUF];

    if (Cnt > MAX_MB_PAKET_DATA) {
    
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return FALSE;
    }

    // Modbus-Befehl aufbauen
    WBuf[0] = GetDeviceIDRemote();
    h = Cnt + 10;
    WBuf[1] = (unsigned char)(h >> 8);
    WBuf[2] = (unsigned char)(h & 0xFF);	// Laenge des Paketes
    WBuf[3] = MB_WRITE;
    ULongToBin(Add, &WBuf[4]);
    
    // Daten in Paket kopieren
    for (i=0; i < Cnt; i++) {

        WBuf[i + 8] = Buf[i];
    }
    MkPaketCrc(WBuf, 8 + Cnt); 
    ret = -1;

    bool bCRCFail = false;

crc_fail:

    if (!Command(WBuf, 10 + Cnt, RBuf, &ret, piError)) {
        
        return FALSE;
    }

    // Mindestens 6 fuer Fehlerpaket
    if (ret <= 4) {
    
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    if (!CheckPaketCrc(RBuf, ret)) {

        if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return FALSE;
    }

    if (RBuf[3] == MB_WRITE && ret == 12) {

        return TRUE;

    } else if (RBuf[3] == (MB_WRITE | 0x80) && ret == 7 && RBuf[4] == 0x22) {

        Sleep(150);
        return TRUE;

    } 

    // Antwort vom Falschen Geraet oder nicht sinnvoll
    *piError = IV_ERROR_ADDRESS;
    return FALSE;
}

// Lesen aus AEIII+ (High-Level) 
// Add -> Byte Adresse
// Cnt -> Anzahl Zeichen (max. 1MByte)
// Buf -> Buffer fuer empfangene Zeichen
int CChannel1_2::ReadMem(unsigned int Add, int Cnt, unsigned char *Buf, int *piError)
{
    int i, j, ac, End = FALSE, ok, MaxB = 32;

    // Fehler loeschen
    *piError = IV_ERROR_SUCCESS;

    // Bis zu 32 Byte pro Paket aus Geraet lesen (anfordern).
    // Daten in Paket zerlegen
    for (i=0; !End; i += MaxB) {

        // Zerlegen
        if ((Cnt - i) < MaxB) {

            ac = Cnt - i;
        } else {

            ac = MaxB;
        }

        // Daten vorhanden?
        if (ac > 0) {

            ok = FALSE;

            // Versuchtes anfragen
            for (j=0; j < 4 && !ok; j++) {

                if (ReadXByte(Add + i, ac, &Buf[i], piError)) {
                    
                    ok = TRUE; 
                } else {
                    
                    // Abbruch durch CallBack!
                    if (*piError == IV_ERROR_USER_CANCEL) {
                        
                        return FALSE;
                    }

                    Sleep(50);
                }
            }
            
            // Fehlgeschlagen
            if (!ok) {

                // Letzten Fehler zurueckgeben
                return FALSE;
            }

        } else {

            End = TRUE;
        }
    }

    return TRUE;
}

// Einfache Anfrage ob Geraet mit Protokollversion 2 erreichbar.
int CChannel1_2::Request(unsigned char deviceIDRemote)
{
    int ret, err;
    unsigned char WBuf[MAX_WBUF];
    unsigned char RBuf[MAX_RBUF];

    // Busnummer einstellen
    if (!SetDeviceIDRemote(deviceIDRemote)) {

        //*piError = IV_ERROR_DEVICE_ID_REMOTE;
        return FALSE;
    }

    // Modbus-Befehl aufbauen
    WBuf[0] = deviceIDRemote;
    WBuf[1] = 0;
    WBuf[2] = 0x0C;		// Laenge des Paketes
    WBuf[3] = MB_READ;
    ULongToBin(0, &WBuf[4]);
    UShortToBin(2, &WBuf[8]);
    MkPaketCrc(WBuf, 10);

    ret = -1;

    if (!Command(WBuf, 12, RBuf, &ret, &err)) {

        return FALSE;
    }

    if (ret <= 4) {
    
        //*piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // Paket pruefen
    if (!CheckPaketCrc(RBuf, ret)) {
    
        //*piError = IV_ERROR_CRC_PACKET;
        return FALSE;
    }

    if (BinToUShort(&RBuf[1]) != (2 + 10)) {
    
        //*piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // auf Fehler pruefen
    if (RBuf[3] != MB_READ || RBuf[0] != GetDeviceIDRemote()) {

        //*piError = IV_ERROR_ADDRESS;
        return FALSE;
    }

    // Ein TimeboyIV unterstuetzt auch das AEIII+ Protokoll auf dieser Ebene.
    // Es sendet auf die Ping-Anfragen (Lese 2 Byte von Add-0x00) jedoch ~~ als Dateninhalt.
    if (RBuf[8] == '~' && RBuf[9] == '~')
    {
        // Es handelt sich um einen TB.
        return FALSE;

    }
    // Ermittelte Protokollversion halten
    m_protocolVersion = 1;

    return TRUE;
}

// X Byte von AEIII+ lesen (Low-Level)
// Add -> Byte Adresse
// Cnt -> Anzahl Zeichen (max. 32 Byte)
// Buf -> Buffer fuer empfangene Zeichen
int CChannel1_2::ReadXByte (unsigned int Add, int Cnt, unsigned char *Buf, int *piError)
{
    int ret, i;
    unsigned char WBuf[MAX_WBUF];
    unsigned char RBuf[MAX_RBUF];

    if (Cnt > MAX_MB_PAKET_DATA) {
        
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return FALSE;
    }

    if (Cnt > 32) {

        GetChannelBase().trace("?");
    }

    // Modbus-Befehl aufbauen
    WBuf[0] = GetDeviceIDRemote();
    WBuf[1] = 0;
    WBuf[2] = 0x0C;		// Laenge des Paketes
    WBuf[3] = MB_READ;
    ULongToBin(Add, &WBuf[4]);
    UShortToBin(Cnt, &WBuf[8]);
    MkPaketCrc(WBuf, 10);
    ret = -1;

    bool bCRCFail = false;

crc_fail:

    if (!Command(WBuf, 12, RBuf, &ret, piError)) {

        return FALSE;
    }

    if (ret <= 4) {
    
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // Paket pruefen
    if (!CheckPaketCrc(RBuf, ret)) {
    
        if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return FALSE;
    }

    if (RBuf[3] == (MB_READ | 0x80) && ret == 7 && RBuf[4] == 0x22) {

        Sleep(150);
    }

    if (BinToUShort(&RBuf[1]) != (Cnt + 10)) {
    
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // auf Fehler pruefen
    if (RBuf[3] != MB_READ || RBuf[0] != GetDeviceIDRemote()) {

        GetChannelBase().trace("Fehler Befehl\n");
        *piError = IV_ERROR_ADDRESS;
        return FALSE;
    }

    // Adresse pruefen
    for (i=0; i < 4; i++) {

        if (RBuf[i+4] != WBuf[i+4]) {

            *piError = IV_ERROR_RECEIVE;
            return FALSE;
        }
    }

    // Daten kopieren
    for (i=0; i < Cnt; i++) {

        Buf[i] = RBuf[i + 8];
    }

    return TRUE;
}

void CChannel1_2::SetCBFunction(int (*pFunction)()) 
{
    m_callBack1 = pFunction;
}

int CChannel1_2::ExeCallBack()
{
    if (m_callBack1) {

        return m_callBack1();

    } else {

        return TRUE;
    }
}

int CChannel1_2::GetProtocolVersion2(short deviceIDRemote, unsigned char detailed, int *piError)
{
    unsigned char Buf[4];

    // Detailierte Version auslesen.
    if (detailed)
    {
        // Version fuer Setup lesen
        if (!ReadMem(MB_GET_VERSION, 2, Buf, piError))
        {
            return 0;
        }
    }
    else 
    {
        if (m_protocolVersion == 1)
        {
            // Es wird angenommen ein AEIII+ mit neuester Com-Version
            Buf[0] = 0x04;
            Buf[1] = 0x02;
        }
        else
        {
            // Es wird angenommen ein PZE Master IV mit neuester Com-Version
            Buf[0] = 0x05;
            Buf[1] = 0x01;
        }
    }

    // Version fuer Kommunikation

    // Ab Firmware 04.01.00 Protokollanpassung (Version 4) und Listen in Tabellenart abhandeln.
    if ((Buf[0] == 7 && Buf[1] == 0) ||
        // Ab Firmware 04.01.00 Protokollanpassung (Version 3)
            (Buf[0] == 6 && Buf[1] == 0) ||
            // Ab Firmware 04.00.23 Listengroesse auf 256kByte erhoeht.
            (Buf[0] == 5 && Buf[1] == 2) ||
            // Ab Firmware 04.00.21 Listen im Hintergrund laden
            (Buf[0] == 5 && Buf[1] == 1) ||
            // Ab Firmware 04.00.19 des PZEMasterIV (Version 2)
            (Buf[0] == 5 && Buf[1] == 0) ||
            // Ab Firmware 03.01.21 wegen auslesen der Firmwareversion
            (Buf[0] == 4 && Buf[1] == 2) ||
            // Ab Firmware 03.00.12 mit Zutrittskontrolle
            (Buf[0] == 4 && Buf[1] == 1) ||
            // Ab Firmware 3.0 mit 16kByte Flash fuer Setupdaten und 16kByte Extended Flash
            (Buf[0] == 4 && Buf[1] == 0) )
    { 
        // Version als Rueckgabe
        return (Buf[0] << 8) | Buf[1];
    }

    *piError = IV_ERROR_VERSION_FIRMWARE;
    return 0;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int CChannel1_2::CheckData(unsigned char deviceIDRemote, int *piError)
{
    unsigned char ucIsData = FALSE;

    // Testen ob Schnittstelle geoeffnet
    if (GetChannelBase().getType() == Channel4Base::Undefined) {

        *piError = IV_ERROR_CHANNEL_CLOSED;
        return -1;
    }

    // Busnummer wechsel, ohne Daten zu loeschen.
    if (!SetDeviceIDRemote(deviceIDRemote)) {

        *piError = IV_ERROR_DEVICE_ID_REMOTE;
        return -1;
    }

    // Abfrage ob Daten vorhanden
    if (!ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError)) {
        
        return -1;
    }

    return ucIsData;
}

// Lesen eines Datensatzes.
// Rueckgabewerte: -1 - Lesen gescheitert
//					0 - Keine Datensaetzte vorhanden
//					1 - Online Datensatz gelesen
//					2 - Wiederholt Online Datensatz gelesen
//					3 - Offline Datensatz gelesen
//					4 - Wiederholt Offline Datensatz gelesen
int CChannel1_2::ReadRecord(unsigned char deviceIDRemote, unsigned char *pucRxBuf, int *piRxCount, int *piError)
{
    if (GetChannelBase().getType() == Channel4Base::Undefined) {
        
        *piError = IV_ERROR_CHANNEL_CLOSED;
        return -1;
    }

    int iResult = -1;
    //Senden des Lese-Datensatz-Befehl
    m_pucTxBuf[0] = deviceIDRemote;			// Bus-Nummer
    m_pucTxBuf[1] = 0x00; 
    m_pucTxBuf[2] = 0x06;				// Befehlslaenge = 6
    m_pucTxBuf[3] = MB_CMD_READ_RECORD;	// Befehl: Lesen eines Datensatzes
    MkPaketCrc(m_pucTxBuf, 4);
    
    bool bEnd = false, bCRCFail = false;

crc_fail:

    for (int i=0; (i < 3) && !bEnd; i++) { // 3 Versuche 

        if(!Command(m_pucTxBuf, 6, pucRxBuf, &iResult, piError)) {
            
            Sleep(50);
            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();

        } else {
        
            bEnd = true;
        }
    }

    // Fehlerhafte versuche
    if (!bEnd) {

        return -1;
    }

    // Paket kleiner als Mindestlaenge?
    if(iResult < 6){

        *piError = IV_ERROR_RECEIVE;
        return -1;
    }
    
    // Pruefsumme OK?
    if(!CheckPaketCrc(pucRxBuf, iResult)){
        
        if (!bCRCFail) { // nach CRC - Fehler nochmals versuchen

            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return -1;
    }
    
    *piRxCount = iResult;

    // Pruefen der Antwort fuer Rueckgabe
    switch(pucRxBuf[3]) {
    case MB_ANSW_NO_RECORD:         // kein Datensatz vorhanden 

        return 0;

    case MB_ANSW_ONLINE_RECORD:     // Online Datensatz 

        return 1;

    case MB_ANSW_OLD_ONLINE_RECORD: // wiederholt gesendeter Online Datensatz 

        return 2;

    case MB_ANSW_OFFLINE_RECORD:    // Offline Datensatz 

        return 3;

    case MB_ANSW_OLD_OFFLINE_RECORD:// wiederholt gesendeter Offline Datensatz 

        return 4;
    };

    // Keine gueltige Antwort auf Datensatzanfrage
    *piError = IV_ERROR_ADDRESS;

    return -1;
}

// Quittieren eines Datensatzes
// Rueckgabewerte:	-1 - Fehler bei Quittierung
//					 0 - Keine Quittung erforderlich
//					 1 - Quittierung erfolgt (online)
//					 2 - Quittierung erfolgt (wiederholte online)
//					 3 - Quittierung erfolgt (offline)
//					 4 - Quittierung erfolgt (wiederholte offline)
int CChannel1_2::QuitRecord(unsigned char deviceIDRemote, unsigned char *pucRxBuf, int *piRxCount, int *piError)
{
    int iResult = FALSE;
    
    if (GetChannelBase().getType() == Channel4Base::Undefined){
        
        *piError = IV_ERROR_CHANNEL_CLOSED;
        return -1;
    }

    // Senden des Lese - Befehls
    m_pucTxBuf[0] = deviceIDRemote;			// Bus-Nummer
    m_pucTxBuf[1] = 0x00; 
    m_pucTxBuf[2] = 0x06;				// Befehlslaenge = 6
    m_pucTxBuf[3] = MB_CMD_ACK_RECORD;	// Befehl: Letzen Datensatz quittieren
    MkPaketCrc(m_pucTxBuf, 4);
    
    bool bEnd = false, bCRCFail = false;

crc_fail:

    for (int i=0; (i < 4) && !bEnd; i++) {

        if(!Command(m_pucTxBuf, 6, pucRxBuf, &iResult, piError)) {
            
            Sleep(50);
            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();

        } else {
            
            bEnd = true;
        }
    }

    // Fehlerhafte Anfrage
    if (!bEnd) {

        return -1;
    }

    // Paket kleiner als Mindestlaenge?
    if(iResult < 6) {
        
        *piError = IV_ERROR_RECEIVE;
        return -1;
    }

    // Pruefsumme OK?
    if(!CheckPaketCrc(pucRxBuf, iResult)) {
        
        if (!bCRCFail) { // Nach CRC - Fehler nochmals anfordern
            
            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return -1;
    }
    
    *piRxCount = iResult;

    switch (pucRxBuf[3]) {
    case 0: // Keine vorherige Datenabholung
    case 2: // Keine vorherige Datenabholung (wiederholt)
        return 0; // Keine Quitierung erforderlich
         
    case MB_ANSW_ACK_ON_RECORD:          // Quittung erhalten (fuer Online-Datensatz)
    case MB_ANSW_TIMEOUT_ACK_RECORD:     // Quittung kam nach Timeout (fuer Online-Datensatz)
        return 1;

    case MB_ANSW_OLD_ACK_ON_RECORD:      // wiederholt gesendete Quittung
    case MB_ANSW_OLD_TIMEOUT_ACK_RECORD: // wiederholte Quittung nach Timeout
        return 2;

    case MB_ANSW_ACK_OFF_RECORD:	     // Quittung erhalten (fuer Offline-Datensatz)
        return 3;

    case MB_ANSW_OLD_ACK_OFF_RECORD:     // wiederholt gesendete Quittung
        return 4;
    }

    // Keine gueltige Antwort auf Datensatzanfrage
    *piError = IV_ERROR_ADDRESS;
    return -1;
}


// Loescht alle Datensaetze
int CChannel1_2::ClearData(unsigned char deviceIDRemote, int *piError)
{
    int iRes;

    if (GetChannelBase().getType() == Channel4Base::Undefined){

        *piError = IV_ERROR_CHANNEL_CLOSED;
        return FALSE;
    }

    // Senden des Lesen-Befehls
    m_pucTxBuf[0] = deviceIDRemote;	// Bus-Nummer
    m_pucTxBuf[1] = 0x00; 
    m_pucTxBuf[2] = 0x06;		// Befehlslaenge = 6
    m_pucTxBuf[3] = MB_CMD_DEL;	// Befehl -Loeschen-

    MkPaketCrc(m_pucTxBuf, 4);

    bool bEnd = false, bCRCFail = false;

crc_fail:

    for (int i=0; (i < 4) && !bEnd; i++) {

        if(!Command(m_pucTxBuf, 6, m_pucRxBuf, &iRes, piError)){
            
            Sleep(50);
            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();
            
        } else {
        
            bEnd = true;
        }
    }

    if (!bEnd) {
    
        return FALSE;
    }

    if(iRes != 6){
        
        // Paket ungleich der Vorgeschriebenen Antwortlaenge?
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // Antwort muss genauso aussehen wie der Befehl, also jedes Byte ueberpruefen
    // Checksumme muss daher nicht ueberprueft werden
    for(int counter=0; counter < iRes; counter++) {
        
        if(m_pucTxBuf[counter] != m_pucRxBuf[counter]){

            if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

                // Nachfolgende Zeichen verwerfen
                GetChannelBase().flush();

                bCRCFail = true;
                goto crc_fail;
            }

            *piError = IV_ERROR_CRC_PACKET;
            return FALSE;
        }
    }

    return TRUE;
}

// Schliesst Relais mit Nummer fuer Zeit.
int CChannel1_2::CloseRelay(unsigned char deviceIDRemote, int iNum, int iTime, int *piError)
{
    int iResult = -1, iError;

    // Senden des Lese - Befehls
    m_pucTxBuf[0] = deviceIDRemote;	// Bus-Nummer
    m_pucTxBuf[1] = 0x00; 
    m_pucTxBuf[2] = 0x0E;		// Befehlslaenge = 14
    m_pucTxBuf[3] = MB_WRITE;	// Befehl: Relais schreiben.
    
    // Adresse fuer Relais
    ULongToBin(MB_RELAY, &m_pucTxBuf[4]);

    m_pucTxBuf[8] = 0x80;           // Relais schliessen
    m_pucTxBuf[8] |= (iNum & 0x0F); // Relaisnummer

    m_pucTxBuf[9] = 0;

    // Zeit die Relais geschlossen wird.
    UShortToBin(iTime, &m_pucTxBuf[10]);
    
    MkPaketCrc(m_pucTxBuf, 12);

    bool bEnd = false, bCRCFail = false;

crc_fail:

    for (int i=0; (i < 4) && !bEnd; i++) {

        if(!Command(m_pucTxBuf, 14, m_pucRxBuf, &iResult, &iError)){
            
            *piError = iError;
            Sleep(50);
            
        } else {
        
            bEnd = true;
        }
    }

    if (!bEnd) {
    
        return FALSE;
    }

    if(iResult != 14){
        
        // Paket ungleich der Vorgeschriebenen Antwortlaenge?
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // CRC - pruefen
    if (!CheckPaketCrc(m_pucRxBuf, iResult)) {

        if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return FALSE;
    }

    return TRUE;
}

// Ermittelt aktuellen Zustand von Relais mit Nummer fuer Zeit.
int CChannel1_2::GetRelayState(unsigned char deviceIDRemote, int iNum, int *piState, int *piTime, int *piError)
{
    if (GetChannelBase().getType() == Channel4Base::Undefined){

        *piError = IV_ERROR_CHANNEL_CLOSED;
        return FALSE;
    }

    int iResult = FALSE;

    // Senden des Lesen-Befehls
    m_pucTxBuf[0] = deviceIDRemote;	// Bus - Nummer
    m_pucTxBuf[1] = 0x00; 
    m_pucTxBuf[2] = 0x0E;		// Befehlslaenge = 14
    m_pucTxBuf[3] = MB_READ;	// Befehl: Relais lesen
    
    // Adresse fuer Relais
    ULongToBin(MB_RELAY, &m_pucTxBuf[4]);

    m_pucTxBuf[8] = (iNum & 0x0F); // Nummer des Relais
    m_pucTxBuf[9] = 0;
    // Dauer
    m_pucTxBuf[10] = 0;
    m_pucTxBuf[11] = 0;

    MkPaketCrc(m_pucTxBuf, 12);

    bool bEnd = false, bCRCFail = false;

crc_fail:

    for (int i=0; (i < 4) && !bEnd; i++) {

        if(!Command(m_pucTxBuf, 14, m_pucRxBuf, &iResult, piError)){
            
            Sleep(50);
            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();
            
        } else {
        
            bEnd = true;
        }
    }

    if (!bEnd) {
    
        return FALSE;
    }

    if(iResult != 14){
        
        // Paket ungleich der Vorgeschriebenen Antwortlaenge?
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // CRC - pruefen
    if (!CheckPaketCrc(m_pucRxBuf, iResult)) {

        if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return FALSE;
    }

    if (m_pucRxBuf[8] & 0x80) {
    
        *piState = 1;
        *piTime = BinToUShort(&m_pucRxBuf[10]);
    } else {
    
        *piState = 0; // nicht gesetzt
        *piTime = 0;  // Zeit nicht relevant
    }

    return TRUE;
}

// Oeffnet Relais mit Nummer.
int CChannel1_2::OpenRelay(unsigned char deviceIDRemote, int iNum, int *piError)
{
    if (GetChannelBase().getType() == Channel4Base::Undefined){

        *piError = IV_ERROR_CHANNEL_CLOSED;
        return FALSE;
    }

    int iResult = FALSE;

    // Senden des Lese - Befehls
    m_pucTxBuf[0] = deviceIDRemote;	// Bus-Nummer
    m_pucTxBuf[1] = 0x00; 
    m_pucTxBuf[2] = 0x0E;		// Befehlslaenge = 14
    m_pucTxBuf[3] = MB_WRITE;	// Befehl: Relais schreiben
    
    // Adresse fuer Relais
    ULongToBin(MB_RELAY, &m_pucTxBuf[4]);

    m_pucTxBuf[8] = (iNum & 0x0F);
    m_pucTxBuf[9] = 0x00;
    m_pucTxBuf[10] = 0x00;
    m_pucTxBuf[11] = 0x00;

    MkPaketCrc(m_pucTxBuf, 12);

    bool bEnd = false, bCRCFail = false;

crc_fail:

    for (int i=0; (i < 4) && !bEnd; i++) {

        if(!Command(m_pucTxBuf, 14, m_pucRxBuf, &iResult, piError)){
            
            Sleep(50);
            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();
            
        } else {
        
            bEnd = true;
        }
    }

    if (!bEnd) {
    
        return FALSE;
    }

    if(iResult != 14){
        
        // Paket ungleich der Vorgeschriebenen Antwortlaenge?
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    // CRC - pruefen
    if (!CheckPaketCrc(m_pucRxBuf, iResult)) {

        if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return FALSE;
    }

    return TRUE;
}

// Daten einer Variablen setzen.
int CChannel1_2::SetGlobVar(unsigned char deviceIDRemote, unsigned char *pucVar, int iType, char *val, int *piError)
{
    if (GetChannelBase().getType() == Channel4Base::Undefined) {

        *piError = IV_ERROR_CHANNEL_CLOSED;
        return FALSE;
    }

    int iResult = FALSE;

    // Laenge ermitteln.
    size_t iValueLength = strlen((const char*)val);

    // Senden des Lesen-Befehls
    m_pucTxBuf[0] = deviceIDRemote;	// Bus-Nummer
    m_pucTxBuf[1] = 0x00; 
    m_pucTxBuf[2] = 0x0C;		// Rahmen Befehlslaenge = 12
    // Maximal 60 Zeichen schreiben.
    if (iValueLength > 60) {
        
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return FALSE;
    }
    iValueLength++; // 0 terminie

    m_pucTxBuf[3] = MB_WRITE;	// Befehl: Globale Variable schreiben.
    
    // Adresse fuer Globale Variable
    ULongToBin(MB_GLOBAL_VAR, &m_pucTxBuf[4]);

    unsigned int iNameLength = 0;
    switch(iType) {
    case 0:
        m_pucTxBuf[8] = (*pucVar & 0x7F); // Nummer der zu lesenden Globalen Variable
        break;
    case 1:
        m_pucTxBuf[8] = 0x80;
        iNameLength = (strlen((const char*)pucVar) & 0x7F) + 1;
        strcpy((char*)&m_pucTxBuf[10], (const char*)pucVar);
        m_pucTxBuf[10 + (iNameLength - 1)] = 0;
        break;
    case 2:
        m_pucTxBuf[8] = 0x80;
        iNameLength = (strlen((const char*)pucVar) & 0x7F) + 1;
        strcpy((char*)&m_pucTxBuf[10], (const char*)pucVar);
        m_pucTxBuf[10 + (iNameLength - 1)] = 0;
        break;
    default:
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return FALSE;
    }

    m_pucTxBuf[9] = 0;

    // Wert in Paket kopieren
    strcpy((char*)&m_pucTxBuf[10 + iNameLength], (const char*)val);
    m_pucTxBuf[10 + iNameLength + (iValueLength - 1)] = 0;

    UShortToBin(m_pucTxBuf[2] + iNameLength + (unsigned short)iValueLength + 2, &m_pucTxBuf[1]);
    MkPaketCrc(m_pucTxBuf, BinToUShort(&m_pucTxBuf[1]) - 2);

    bool bEnd = false, bCRCFail = false;

crc_fail:

    for (int i=0; (i < 4) && !bEnd; i++) {

        if(!Command(m_pucTxBuf, BinToUShort(&m_pucTxBuf[1]), m_pucRxBuf, &iResult, piError)){
            
            Sleep(50);
            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();
            
        } else {
        
            bEnd = true;
        }
    }

    if (!bEnd) {
    
        return FALSE;
    }

    if(iResult != BinToUShort(&m_pucRxBuf[1])) {
        
        // Paket ungleich der Vorgeschriebenen Antwortlaenge?
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    if(BinToUShort(&m_pucRxBuf[1]) == 6) {
        
        // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
        // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
        if (m_pucRxBuf[3] & 0x80) {
            // Fehler liegt bei Wert
            *piError = IV_ERROR_VARIABLE_VALUE;
        } else {
            // Variable nicht bekannt
            *piError = IV_ERROR_VARIABLE_UNDEFINED;
        }

        return FALSE;
    }

    // CRC - pruefen
    if (!CheckPaketCrc(m_pucRxBuf, iResult)) {

        if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return FALSE;
    }

    return TRUE;
}

// Daten einer Globalen Variablen lesen.
int CChannel1_2::GetGlobVar(unsigned char deviceIDRemote, unsigned char *pucVar, int iType, char *val, int iValueLength, int *piError)
{
    if (GetChannelBase().getType() == Channel4Base::Undefined){

        *piError = IV_ERROR_CHANNEL_CLOSED;
        return FALSE;
    }

    // muss mindestens ein Zeichen anfordern.
    if ((iValueLength - 1) <= 0) {
    
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return FALSE;
    } 

    int iResult = FALSE;

    // Senden des Lesen-Befehls
    m_pucTxBuf[0] = deviceIDRemote;	// Bus - Nummer
    m_pucTxBuf[1] = 0x00; 
    m_pucTxBuf[2] = 0x0C;		// Rahmen Befehlslaenge = 12
    m_pucTxBuf[3] = MB_READ;	// Befehl: Globale Variable lesen.
    
    // Adresse fuer Globale Variable
    ULongToBin(MB_GLOBAL_VAR, &m_pucTxBuf[4]);

    unsigned int iNameLength = 0;
    switch(iType) {
    case 0:
        m_pucTxBuf[8] = (*pucVar & 0x7F); // Nummer der zu lesenden Globalen Variable
        break;
    case 1:
        m_pucTxBuf[8] = 0x80;
        iNameLength = (strlen((const char*)pucVar) & 0x7F) + 1;
        strcpy((char*)&m_pucTxBuf[10], (const char*)pucVar);
        m_pucTxBuf[10 + (iNameLength - 1)] = 0;
        break;
    case 2:
        m_pucTxBuf[8] = 0x80;
        iNameLength = (strlen((const char*)pucVar) & 0x7F) + 1;
        strcpy((char*)&m_pucTxBuf[10], (const char*)pucVar);
        m_pucTxBuf[10 + (iNameLength - 1)] = 0;
        break;
    default:
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return FALSE;
    }

    // Laenge des zur Verfuegung stehenden Puffer - Nullterminie
    m_pucTxBuf[9] = (iValueLength > 60) ? 59 : (iValueLength - 1); 

    UShortToBin(m_pucTxBuf[2] + iNameLength + 2, &m_pucTxBuf[1]);
    MkPaketCrc(m_pucTxBuf, BinToUShort(&m_pucTxBuf[1]) - 2);

    bool bEnd = false, bCRCFail = false;

crc_fail:

    for (int i=0; (i < 4) && !bEnd; i++) {

        if(!Command(m_pucTxBuf, BinToUShort(&m_pucTxBuf[1]), m_pucRxBuf, &iResult, piError)){
            
            Sleep(50);
            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();
            
        } else {
        
            bEnd = true;
        }
    }

    if (!bEnd) {
    
        return FALSE;
    }

    if(iResult != BinToUShort(&m_pucRxBuf[1])){
        
        // Paket ungleich der Vorgeschriebenen Antwortlaenge?
        *piError = IV_ERROR_RECEIVE;
        return FALSE;
    }

    if(BinToUShort(&m_pucRxBuf[1]) == 6) {
        
        // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
        // und wenn der Wert nicht gelesen werden konnte wird das hoechste bit gesetzt.
        if (m_pucRxBuf[3] & 0x80) {
            // Fehler liegt bei Wert
            *piError = IV_ERROR_VARIABLE_VALUE;
        } else {
            // Variable nicht bekannt
            *piError = IV_ERROR_VARIABLE_UNDEFINED;
        }

        return FALSE;
    }

    // CRC - pruefen
    if (!CheckPaketCrc(m_pucRxBuf, iResult)) {

        if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

            // Nachfolgende Zeichen verwerfen
            GetChannelBase().flush();

            bCRCFail = true;
            goto crc_fail;
        }

        *piError = IV_ERROR_CRC_PACKET;
        return FALSE;
    }


    // Pruefe die zu kopierende Laenge.
    if (m_pucRxBuf[9] > (iValueLength - 1)) {
    
        strncpy(val, (const char*)&m_pucRxBuf[10 + iNameLength], (iValueLength - 1));
        val[iValueLength - 1] = 0;

        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return FALSE;

    } else {

        strncpy(val, (const char*)&m_pucRxBuf[10 + iNameLength], m_pucRxBuf[9]);
        val[m_pucRxBuf[9]] = 0;
    }

    return TRUE;
}

int CChannel1_2::GetHardwareInfo(short deviceIDRemote, StructHardwareInfo &hwInfo, int *piError)
{
    (void)deviceIDRemote;
    // Struktur initialisieren
    memset(&hwInfo, 0, sizeof(StructHardwareInfo));


    // Geraetetyp ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FCB, 1, &hwInfo.typeDevice, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeDevice == 0xFF) {

        hwInfo.typeDevice = 0; // PZE-MasterIV
    }

    // Boardtyp ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FCD, 1, &hwInfo.typeBoard, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeBoard == 0xFF) {

        hwInfo.typeBoard = 0x13; // Baugruppe 1.3
    }
    
    // RS485 ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FCA, 1, &hwInfo.componentRS485_0, piError)) {
        
        return FALSE;
    }
    if (hwInfo.componentRS485_0 == 0xFF) {

        hwInfo.componentRS485_0 = 0; // Nein
    }

    // RS485 und TCPIP ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FD2, 1, &hwInfo.typeTCPIP, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeTCPIP == 0xFF) {

        hwInfo.typeTCPIP = 0;
    }
    // RS485 und TCPIP pruefen
    if (hwInfo.componentRS485_0 == 0 && hwInfo.typeTCPIP == 0) {
        // Wenn keines von beiden, dann wird RS485 angenommen.
        hwInfo.componentRS485_0 = 1;
    }
    
    // ModemMC35i ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FD3, 1, &hwInfo.typeModemMC, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeModemMC == 0xFF) {

        hwInfo.typeModemMC = 0;
    }
    
    // RS485 (intern) ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FC9, 1, &hwInfo.componentRS485_1, piError)) {
        
        return FALSE;
    }
    if (hwInfo.componentRS485_1 == 0xFF) {

        hwInfo.componentRS485_1 = 1;
    }

    // Displaytyp ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FC8, 1, &hwInfo.typeDisplay, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeDisplay == 0xFF) {

        hwInfo.typeDisplay = 1; // Lehner Klein
    }

    // Tastaturtyp ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FC6, 1, &hwInfo.typeKeypad, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeKeypad == 0xFF) {

        hwInfo.typeKeypad = 1; // 9 Tasten
    }
    
    // Transpondertyp ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FD5, 1, &hwInfo.typeTransponder, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeTransponder == 0xFF) {

        hwInfo.typeTransponder = 0; // Kein Transponder
    }

    // Typ ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FC7, 1, &hwInfo.typeMMC, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeMMC == 0xFF) {

        hwInfo.typeMMC = 0; // Keine MMC
    }
    
    // Typ ermitteln
    if (!ReadMem(MB_START_EEP + 0x0F99, 1, &hwInfo.typeRadio, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeRadio == 0xFF) {

        hwInfo.typeRadio = 0;
    }
    
    // Typ ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FBC, 1, &hwInfo.typeGPS, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeGPS == 0xFF) {

        hwInfo.typeGPS = 0;
    }
    
    // Komponende ermitteln
    if (!ReadMem(MB_START_EEP + 0x0FBD, 1, &hwInfo.componentUSB, piError)) {
        
        return FALSE;
    }
    if (hwInfo.componentUSB == 0xFF) {

        hwInfo.componentUSB = 0;
    }
    
    // Typ ermitteln
    if (!ReadMem(MB_START_EEP + 0x0F98, 1, &hwInfo.typeDCF77, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeDCF77 == 0xFF) {

        hwInfo.typeDCF77 = 0;
    }
    
    // Komponende ermitteln
    if (!ReadMem(MB_START_EEP + 0x0EDC, 4, (unsigned char*)&hwInfo.componentInAnalog, piError)) {
        
        return FALSE;
    }
    if (hwInfo.componentInAnalog == 0xFFFFFFFF) {

        hwInfo.componentInAnalog = 0;
    }
    
    // Komponende ermitteln
    if (!ReadMem(MB_START_EEP + 0x0F90, 4, (unsigned char*)&hwInfo.componentInDigital, piError)) {
        
        return FALSE;
    }
    if (hwInfo.componentInDigital == 0xFFFFFFFF) {

        hwInfo.componentInDigital = 0;
    }
    
    // Komponende ermitteln
    if (!ReadMem(MB_START_EEP + 0x0F94, 4, (unsigned char*)&hwInfo.componentOutDigital, piError)) {
        
        return FALSE;
    }
    if (hwInfo.componentOutDigital == 0xFFFFFFFF) {

        hwInfo.componentOutDigital = 0;
    }
    
    // Typ ermitteln
    if (!ReadMem(MB_START_EEP + 0x0EDB, 1, &hwInfo.typeFingerprint, piError)) {
        
        return FALSE;
    }
    if (hwInfo.typeFingerprint == 0xFF) {

        hwInfo.typeFingerprint = 0;
    }

    // Typ ermitteln
    if (!ReadMem(MB_START_EEP + 0x0EDA, 1, &hwInfo.typeExternalReader, piError)) {
        
        return FALSE;
    }
    // Wenn ZK Option vorhanden
    if (hwInfo.componentRS485_1) {
        // aber der Externe Lesertyp undefiniert ist,
        // dann wird von TS-Lesern ausgegangen.
        if (hwInfo.typeExternalReader == 0xFF) {

            hwInfo.typeExternalReader = 0x01;
        }
    } else {

        // Keine Externen leser
        hwInfo.typeExternalReader = 0x00;
    }
    // BUGFIX
    // Eine Firmware die die Serien TS und Protect unterstützt setzt als unterstützte Serien
    // TS und Protect obwohl es so sein sollte, dass die angesprochene Serie eingetragen wird,
    // so dass der Wert der Hardwareinfo eindeutig ist.
    // Werden beide Flags vorgefunden oder Protect alleine wird von TS alleine ausgegangen.
    if (hwInfo.typeExternalReader & 0x04) {

        hwInfo.typeExternalReader = 0x01;
    }


    return TRUE;
}

int CChannel1_2::GetTables( unsigned char ucList, int *piError )
{
    unsigned int pDat, pFeld, pList[MAX_DATENSATZ_BESCHREIBUNG], pListEnd;
    int i, j;

    // EEPROM lesen 
    if (!ReadMem(MB_START_EEP, EEP_SIZE_SU, m_file.m_eep, piError))
    {
        return 0;
    }

    // Start und Ende der Datensatzbeschreibungen lesen
    ChangeOrderLong(&m_file.m_eep[P_DAT_STRUCT]);
    StartDatensatzB = *(unsigned int*)(&m_file.m_eep[P_DAT_STRUCT]);

    ChangeOrderLong(&m_file.m_eep[P_LIST_STRUCT]);
    EndDatensatzB = *(unsigned int*)(&m_file.m_eep[P_LIST_STRUCT]);

    if (StartDatensatzB >=  MAX_FLASH_MEM || EndDatensatzB >= MAX_FLASH_MEM)
    {			
        *piError = IV_ERROR_CHANNEL_COMMUNICATION;
        return 0;
    }

    if (ucList)
    {
        //Pointer der Listenbeschreibungen lesen (Ende Datensatzbeschreibung ist anfang Listenb.
        if (!ReadMem(EndDatensatzB, MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int), &m_file.m_flash[EndDatensatzB - FLASH_KONFIG_START], piError))
        {	
            return 0;
        }

        // Pointer extrahieren und Anzahl Listenb. ermitteln
        for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++)
        {
            ChangeOrderLong(&m_file.m_flash[EndDatensatzB - FLASH_KONFIG_START + i * sizeof(unsigned int)]);
            pList[i] = *(unsigned int*)(&m_file.m_flash[EndDatensatzB - FLASH_KONFIG_START + i *sizeof(unsigned int)]);

            if (pList[i] == 0ul)
            {					
                MaxListB = i; 
                i = MAX_DATENSATZ_BESCHREIBUNG; 
                continue;
            }
        }

        // Ende Listenbereich angenaehert berechnen (mit Reserve)
        pListEnd = pList[MaxListB - 1] + sizeof(DatensatzBeschreibung[0]) + 
            (sizeof(DatensatzFeld[0][0]) + 
            sizeof(unsigned int)) * (MAX_DATENSATZ_BESCHREIBUNG + 1);

        // Rest der Listenbeschreibung aus AEIII+ lesen
        pDat = EndDatensatzB + MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int);

        if (!ReadMem(pDat, pListEnd - pDat, &m_file.m_flash[pDat - FLASH_KONFIG_START], piError))
        {				
            return 0;
        }

        // Listenbeschreibungen auswerten
        for (i=0; i < MaxListB; i++)
        {
            CpyFromKonfigBuf((unsigned char*)&ListenBeschreibung[i], m_file.m_flash, pList[i], sizeof(ListenBeschreibung[i]));
            for (j=0; j < ListenBeschreibung[i].Felder; j++)
            { 
                // Feldbeschreibung einlesen
                ChangeOrderLong  (&m_file.m_flash[ChangeOrderULong(ListenBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(unsigned int)]);
                pFeld = *(unsigned int*)(&m_file.m_flash[ChangeOrderULong(ListenBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(unsigned int)]);

                //					ASSERT (pFeld + sizeof(DatensatzFeld[i][j]) < pListEnd);

                CpyFromKonfigBuf((unsigned char*)&ListenFeld[i][j], m_file.m_flash, pFeld, sizeof(ListenFeld[i][j]));
            }
        }
    }
    else
    {
        // Datensatzbeschreibungen lesen
        if (!ReadMem(StartDatensatzB, EndDatensatzB - StartDatensatzB, &m_file.m_flash[StartDatensatzB - FLASH_KONFIG_START], piError))
        {				
            return 0;
        }

        // Datensatzbeschreibung auswerten
        for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++)
        {
            ChangeOrderLong(&m_file.m_flash[StartDatensatzB - FLASH_KONFIG_START + i * sizeof(unsigned int)]);
            pDat = *(unsigned int*)(&m_file.m_flash[StartDatensatzB - FLASH_KONFIG_START + i * sizeof(unsigned int)]);

            if (pDat == 0ul)
            {					
                MaxDatB = i; 
                i = MAX_DATENSATZ_BESCHREIBUNG; 
                continue;
            }

            CpyFromKonfigBuf((unsigned char*)&DatensatzBeschreibung[i], m_file.m_flash, pDat, sizeof(DatensatzBeschreibung[i]));

            // Feldbeschreibung einlesen
            for (j=0; j < DatensatzBeschreibung[i].Felder; j++)
            { 				
                ChangeOrderLong(&m_file.m_flash[ChangeOrderULong(DatensatzBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(unsigned int)]);
                pFeld = *(unsigned int*)(&m_file.m_flash[ChangeOrderULong(DatensatzBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(unsigned int)]);

                CpyFromKonfigBuf((unsigned char*)&DatensatzFeld[i][j], m_file.m_flash, pFeld, sizeof(DatensatzFeld[i][j]));
            }
        }
    }		

    return 1;
}

int CChannel1_2::GetListName(unsigned char ucList, char *name, int *piError)
{
    unsigned int add;
    StructDatensatzBeschreibung dsb;

    // Die Addresse auf die Listenbeschreibungen aus dem EEPROM-Bereich lesen 
    if (!ReadMem(MB_START_EEP + P_LIST_STRUCT, sizeof(unsigned int), (unsigned char*)&add, piError))
    {
        return 0;
    }
    ChangeOrderLong((unsigned char*)&add);

    // Die Addresse auf die Listenbeschreibung lesen.
    if (!ReadMem(add + ucList * sizeof(unsigned int), sizeof(unsigned int), (unsigned char*)&add, piError))
    {	
        return 0;
    }
    ChangeOrderLong((unsigned char*)&add);


    // Wenn die Rueckgabe FALSE und Error = IV_ERROR_SUCCESS, dann war die Liste nicht vorhanden.
    *piError = IV_ERROR_SUCCESS;

    if (add)
    {
        // Die Listenbeschreibung lesen.
        if (!ReadMem(add, sizeof(StructDatensatzBeschreibung), (unsigned char*)&dsb, piError))
        {		
            return 0;
        }

        strcpy(name, (char*)dsb.Name);
        return 1;
    }

    return 0;
}

int CChannel1_2::WriteSetupData(char *szFileName, int *piError)
{
    int i, fileError;
    char ext[5];
    memset(ext, 0, sizeof(ext));
    strcpy(ext, &szFileName[strlen(szFileName) - 4]);
    for (i=0; ext[i] && i < 4; i++) {
        ext[i] = tolower(ext[i]);
    }
    if (strcmp(ext, ".aes") != 0 ) {
        // Die Anwendung kann ueber den gewaehlten Weg, nur Dateien vom Typ (*.aes) oeffnen.
        *piError = IV_ERROR_FILE_EXTENSION;
        return 0;
    }

    // Reset der StudioFile-Klasse
    m_file.SetDefault();

    // Laden der Dateidaten. Keine Beachtung der Studioversion.
    if (!m_file.Load(szFileName, &fileError, FILESTUDIOHEADER_IGNORE_MIN_STUDIO_VERSION | FILESTUDIOHEADER_IGNORE_SHOULD_STUDIO_VERSION)) {
        switch (fileError) {
        case -3:
            *piError = IV_ERROR_FILE_READ;
            break;
        case -7: // Mindestversion DLL wird nicht eingehalten.
            *piError = IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
            break;
        case -8: // Sollteversion DLL
            *piError = IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION;
            break;
        case -9: // Format der Setupdatei
            *piError = IV_ERROR_FILE_FORMAT;
            break;
        case -4: // CRC in Dateidaten
        case -10: case -11: case -12: case -13: case -14: case -15: case -16: case -17: case -18:
            *piError = IV_ERROR_FILE_CRC;
            break;
        default:
            *piError = IV_ERROR_FILE_OPEN;
        }
        return 0;
    }

    if (m_file.IsHeader() && (m_file.GetHeader().minFirmware1Version != 0 || m_file.GetHeader().minFirmware2Version != 0)) {
        // Die Firmware wird durch diese Setupdatei nicht unterstuetzt.
        *piError = IV_ERROR_VERSION_SETUPFILE;
        return 0;
    }

    // Gerätetyp der Setupdatei ist weder AEIII+ noch PZE-MasterIV
    if (m_file.m_eep[P_DEVICE_TYPE] != 0 && m_file.m_eep[P_DEVICE_TYPE] != 1) {
        *piError = IV_ERROR_DEVICE_WRONG;
        return 0;
    }

    // Kommunikationsversionsnummer
    int commVersion = 0;

    // Bei AEIII+ zusaetzliche Pruefung wegen Befehlsunterstuetzung notwendig
    commVersion = GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError);
    switch (commVersion)
    {
    case 0x0400:
    case 0x0401:
    case 0x0402:
    case 0x0500:
    case 0x0501:
    case 0x0502:
        break;

    default:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;
    }

    if (commVersion != 0x0400 && commVersion != 0x0401)	{
        char version[32];
        memset(version, 0, sizeof(version));
        // Firmwareversion lesen
        if (!GetVersionFirmware(version, piError))
        {
            return 0;
        }

        CDFComUtils util;
        // Firmwareversion pruefen
        if (!util.CheckVersionFirmware(version))
        {
            *piError = IV_ERROR_VERSION_FIRMWARE;
            return 0;
        }
    }

    unsigned char ucIsData = 0, temp = 1; // fuer Bedienstopp

    // Abfrage ob Daten vorhanden. 
    // (wegen alter FW, nicht mit checkData ausgefuehrt.)
    if (!ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError)) {			
        return 0;
    }
    // Abbruch wenn Daten im Geraet vorliegen
    if (ucIsData) {
        *piError = IV_ERROR_DATA_AVAILABLE;
        return 0;
    }

    // Setupgroesse ermitteln
    unsigned int ulSetupSize = GetFlashSetupSize();
    // Vorhandene zur moeglichen Groesse pruefen
    if (ulSetupSize > FLASH_SIZE_SU) {
        *piError = IV_ERROR_FILE_OPEN;
        return 0;
    }
    // Bedienstopp setzen
    if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {
        return 0;
    }

    // Groesse der Flashdaten in EepMem eintragen
    *(unsigned int*)&m_file.m_eep[P_FLASH_SIZE_SU] = ulSetupSize;
    ChangeOrderLong(&m_file.m_eep[P_FLASH_SIZE_SU]);
    // Groesse der MDE - Eingabeketten in EepMem loeschen.
    *(unsigned int*)&m_file.m_eep[P_MDE_EK_SIZE] = 0xFFFFFFFF;
    ChangeOrderLong(&m_file.m_eep[P_MDE_EK_SIZE]);
    // Groesse der Zutrittskontroll - Eingabeketten in EepMem loeschen.
    *(unsigned int*)&m_file.m_eep[P_ENTRANCE_EK_SIZE] = 0xFFFFFFFF;
    ChangeOrderLong(&m_file.m_eep[P_ENTRANCE_EK_SIZE]);
    // Groesse der MDE - Eingabeketten in EepMem eintragen.
    *(unsigned int*)&m_file.m_eep[P_MDE_EK_SIZE] = GetMDEEKFlashSize();
    ChangeOrderLong(&m_file.m_eep[P_MDE_EK_SIZE]);

    if (commVersion != 0x0400) {
        // Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
        *(unsigned int*)&m_file.m_eep[P_ENTRANCE_EK_SIZE] = GetEntranceEKFlashSize();
        ChangeOrderLong(&m_file.m_eep[P_ENTRANCE_EK_SIZE]);
    }

#define MB_DATABLOCKSIZE 32

    // EEPROM - schreiben 
    if (!WriteMem(MB_START_EEP, EEP_SIZE_SU, m_file.m_eep, piError, MB_DATABLOCKSIZE)) {
        return 0;
    }
    // Flash - schreiben
    if (!WriteMem(FLASH_KONFIG_START, ulSetupSize, m_file.m_flash, piError, MB_DATABLOCKSIZE)) {
        return 0;
    }
    // Konfigurationsdaten - schreiben
    if (!WriteMem(MB_START_CONFIG + 2, CONFIG_SIZE_SU - 2, &m_file.m_configMem[2], piError, MB_DATABLOCKSIZE)) {
        return 0;
    }

    // IOModule und MDE ab Version 3.0
    StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;
    // IOModul - Konfiguration - schreiben
    if (!WriteMem(MB_START_IOMODULE, sizeof(StructIOModule), &m_file.m_flashExtended[extBegin.offsetIOModule], piError, MB_DATABLOCKSIZE)) {
        return 0;
    }
    // IOModul - DigitalMDE - schreiben
    for (i=0; i < 8; i++) {
        if (!WriteMem(MB_DIGITAL_IO + i * sizeof(StructDigitalInput), sizeof(StructDigitalInput), 
            &m_file.m_flashExtended[extBegin.offsetIOModule8DigitalInputs + i * sizeof(StructDigitalInput)], piError, MB_DATABLOCKSIZE))	{
                return 0;
        }
    }

    // Interne MDE - DigitalMDE - schreiben
    for (i=0; i < 4; i++) {
        if (!WriteMem(MB_DIGITAL_MDE + i * sizeof(StructDigitalInput), sizeof(StructDigitalInput), 
            &m_file.m_flashExtended[extBegin.offsetIntern4DigitalInputs + i * sizeof(StructDigitalInput)], piError, MB_DATABLOCKSIZE)) {
                return 0;
        }
    }

    // MDE - Eingabeketten - schreiben
    // Bei 4.0 kommt erst 1Byte Anzahl im Array
    // Ab 4.1 kommen erst 4Byte Anzahl im Array
    unsigned int ulSize = 0;
    if (commVersion == 0x0400) {
        ulSize = GetMDEEKFlashSize_0400();
    } else {				
        ulSize = GetMDEEKFlashSize();
    }

    if (!WriteMem(MB_MDE_EK, ulSize, &m_file.m_flashExtended[extBegin.offsetEventChains], piError, MB_DATABLOCKSIZE)) {
        return 0;
    }

    // Zutrittskontrolle
    if (commVersion != 0x0400) {
        // Zutrittskontroll - Eingabeketten - schreiben
        if (!WriteMem(MB_ENTRANCE_EK, GetEntranceEKFlashSize(), 
            &m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV1], piError, MB_DATABLOCKSIZE)) {
                return 0;
        }
    }


    temp = 0; // Bedienstopp ruecksetzen
    if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {
        return 0;
    }

    // Aktuelle Setup - Daten wurden erst durch zuruecksetzen von Bedienstopp uebernommen!

    // Standardwerte fuer Globale Variable
    if (commVersion != 0x0400) {
        int i;
        StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

        // GV setzen?
        if (extBegin.offsetGlobalVariableValues) {			
            unsigned char ucNum;
            // Struktur lesen
            StructValueGVx ValueGVx = *(StructValueGVx*)&m_file.m_flashExtended[extBegin.offsetGlobalVariableValues];

            // Zeit fuer Umschaltung aus Bedienstopp
            Sleep(200);

            for (i=0; i < 8; i++) {
                if (ValueGVx.ucFlags & (1 << i)) {					
                    ucNum = 1 + i;
                    if (!SetGlobVar(GetDeviceIDRemote(), &ucNum, 0, ValueGVx.szValue[i], piError)) {						
                        return 0;
                    }
                }
            }
        }	
    }

    // Standardwerte fuer Systemvariable
    // Ab 04.00.23
    if (commVersion == 0x0502) {
        //############################
        // 2005-09-19: Erweiterte Struktur fuer Parameter
        if (((m_file.m_eep[P_SETUP_VERSION    ] == 3) &&
            (m_file.m_eep[P_SETUP_VERSION + 1] == 1) &&
            ((m_file.m_eep[P_SETUP_VERSION + 2] > 34) ||
            (m_file.m_eep[P_SETUP_VERSION + 2] == 34) &&
            (m_file.m_eep[P_SETUP_VERSION + 3] >  5)) ) ||

            ((m_file.m_eep[P_SETUP_VERSION   ] == 4) &&
            (m_file.m_eep[P_SETUP_VERSION + 1] == 1) &&
            ((m_file.m_eep[P_SETUP_VERSION + 2] > 0) || 
            (m_file.m_eep[P_SETUP_VERSION + 2] == 0) &&
            (m_file.m_eep[P_SETUP_VERSION + 3] > 2))) )
        {
            StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

            // Liegt die Struktur vor?
            if (extBegin.offsetExtendedStruct) {
                StructExtended extParam = *(StructExtended*)&m_file.m_flashExtended[extBegin.offsetExtendedStruct];

                unsigned char val[32], name[64];
                // Buchungswiederholsperrdauer
                sprintf((char*)name, "_SERVICE.DBLINPUT");
                sprintf((char*)val, "%i", extParam.doubleInputTime);
                // 
                if (!SetGlobVar(GetDeviceIDRemote(), name, 2, (char*)val, piError))	{					
                    if (// Fehler liegt bei Wert
                        *piError != IV_ERROR_VARIABLE_VALUE &&
                        // Variable nicht bekannt
                        *piError != IV_ERROR_VARIABLE_UNDEFINED)
                        return 0;
                }
                // Abblendedauer fuer Bestaetigungsfeld
                sprintf((char*)name, "_SERVICE.DEMANDTIMER");
                sprintf((char*)val, "%i", extParam.confirmationTime);
                // 
                if (!SetGlobVar(GetDeviceIDRemote(), name, 2, (char*)val, piError)) {					
                    if (// Fehler liegt bei Wert
                        *piError != IV_ERROR_VARIABLE_VALUE &&
                        // Variable nicht bekannt
                        *piError != IV_ERROR_VARIABLE_UNDEFINED)
                        return 0;
                }
            }
        }
    }

    return 1;
}

int CChannel1_2::ReadSetupData(char *szFileName, int *piError)
{
    int i;

    char ext[_MAX_EXT];          // 256
#ifdef _MSC_VER
    char drive[_MAX_DRIVE];      // 3
    char dir[_MAX_DIR];          // 256
    char fname[_MAX_FNAME];      // 256

    // Dateiendung pruefen
    _splitpath(szFileName, drive, dir, fname, ext);
#else
    strcpy(ext, &szFileName[strlen(szFileName) - 4]);
#endif
    for (i=0; ext[i] && i < 4; i++)
    {
        ext[i] = tolower(ext[i]);
    }
    if (strcmp(ext, ".aes") != 0 && strcmp(ext, ".ael") != 0)
    {
        // Die Anwendung kann ueber den gewaehlten Weg, nur Dateien vom Typ (*.aes; *.ael) oeffnen.
        *piError = IV_ERROR_FILE_EXTENSION;
        return FALSE;
    }

    int res = 0;

    // Speicher initialisieren
    m_file.SetDefault();

    if (strcmp(ext, ".ael") == 0)
    {
        // Nicht gueltige Endung der Setupdatei.
        *piError = IV_ERROR_FILE_EXTENSION;
        return 0;
    }
    // Kommunikationsversionsnummer
    int commVersion = 0;

    // Bei AEIII+ zusaetzliche Pruefung wegen Befehlsunterstuetzung notwendig
    commVersion = GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError);
    switch (commVersion)
    {
    case 0x0400:
    case 0x0401:
    case 0x0402:
    case 0x0500:
    case 0x0501:
    case 0x0502:
        break;

    default:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;
    }

    if (commVersion != 0x0400 && commVersion != 0x0401)	{
        char version[32];
        memset(version, 0, sizeof(version));
        // Firmwareversion lesen
        if (!GetVersionFirmware(version, piError))
        {
            return 0;
        }

        CDFComUtils util;
        // Firmwareversion pruefen
        if (!util.CheckVersionFirmware(version))
        {
            *piError = IV_ERROR_VERSION_FIRMWARE;
            return 0;
        }
    }

    unsigned char temp = 1; // fuer Bedienstopp
    unsigned int ulTmp, ulSize = FLASH_SIZE_SU;

    // Bedienstopp setzen.
    if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
    {
        return 0;
    }

    // EEPROM - lesen 
    if (!ReadMem(MB_START_EEP, EEP_SIZE_SU, m_file.m_eep, piError))
    {
        return 0;
    }

    // Groesse der Setupdaten bereits im Eep?
    ulTmp = *(unsigned int*)&m_file.m_eep[P_FLASH_SIZE_SU];
    ChangeOrderLong((unsigned char*)&ulTmp);
    // Liegt im moeglichen Bereich?
    if (!(ulTmp > FLASH_SIZE_SU) && !(ulTmp == 0))
    {
        ulSize = ulTmp;
    }

    if (ulSize > FLASH_SIZE_SU)
    {
        *piError = IV_ERROR_FILE_OPEN;
        return 0;
    }

    // Flashbereich - lesen
    if (!ReadMem(FLASH_KONFIG_START, ulSize, m_file.m_flash, piError))
    {
        return 0;
    }

    // Konfiguration - lesen
    if (!ReadMem(MB_START_CONFIG, CONFIG_SIZE_SU, m_file.m_configMem, piError))
    {
        return 0;
    }

    StructExtendedFlash &extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;
    // Flag fuer gelesene Setupdaten setzen
    extBegin.flags = EXTF_DOWNLOAD;

    //*(TExtendedFlash*)m_file.m_flashExtended = extBegin;

    // Struktur fuer IOModul eintragen.
    extBegin.offsetIOModule = extBegin.offsetFreeScope;

    // IOModul - Konfiguration - lesen
    if (!ReadMem(MB_START_IOMODULE, sizeof(StructIOModule), &m_file.m_flashExtended[extBegin.offsetIOModule], piError))
    {
        return 0;
    }

    // Auf naechsten freien Speicher
    extBegin.offsetFreeScope += sizeof(StructIOModule);

    // Strukturen fuer Digitale Eingaenge des IOModules eintragen.
    extBegin.offsetIOModule8DigitalInputs = extBegin.offsetFreeScope;

    // IOModul - DigitalMDE - lesen
    for (i=0; i < 8; i++)
    {
        if (!ReadMem(MB_DIGITAL_IO + i * sizeof(StructDigitalInput), sizeof(StructDigitalInput), 
            &m_file.m_flashExtended[extBegin.offsetIOModule8DigitalInputs + i * sizeof(StructDigitalInput)], piError))
        {
            return 0;
        }
    }

    // Auf naechsten freien Speicher
    extBegin.offsetFreeScope += 8 * sizeof(StructDigitalInput);

    // Strukturen fuer Digitale Eingaenge der MDE eintragen.
    extBegin.offsetIntern4DigitalInputs = extBegin.offsetFreeScope;

    // Interne MDE - DigitalMDE - lesen
    for (i=0; i < 4; i++)
    {
        if (!ReadMem(MB_DIGITAL_MDE + i * sizeof(StructDigitalInput), sizeof(StructDigitalInput), 
            &m_file.m_flashExtended[extBegin.offsetIntern4DigitalInputs + i * sizeof(StructDigitalInput)], piError))
        {
            return 0;
        }
    }

    // Auf naechsten freien Speicher
    extBegin.offsetFreeScope += 4 * sizeof(StructDigitalInput);


    // Strukturen der Eingabeketten von MDE eintragen
    extBegin.offsetEventChains = extBegin.offsetFreeScope;

    // Groesse der MDE - Eingabeketten bereits im Eep?
    ulTmp = *(unsigned int*)&m_file.m_eep[P_MDE_EK_SIZE];
    ChangeOrderLong((unsigned char*)&ulTmp);
    // Liegt im moeglichen Bereich?
    if (!(ulTmp > EXTENDED_SIZE_SU) && !(ulTmp == 0))
    {
        ulSize = ulTmp;
    }
    else
    {		
        ulSize = EXTENDED_SIZE_SU - extBegin.offsetEventChains;
    }


    // MDE - Eingabeketten - lesen
    // Bei 4.0 kommt erst 1Byte Anzahl im Array
    // Ab 4.1 kommen erst 4Byte Anzahl im Array
    if (!ReadMem(MB_MDE_EK, ulSize, &m_file.m_flashExtended[extBegin.offsetEventChains], piError))
    {
        return 0;
    }

    if (commVersion != 0x0400)
    {
        // Auf naechsten freien Speicher
        extBegin.offsetFreeScope += GetMDEEKFlashSize();

        // Dieses sollte nicht sein!, fehlende MDE - Groesse.
        if (ulSize == (EXTENDED_SIZE_SU - extBegin.offsetEventChains))
        {
            *piError = IV_ERROR_FILE_FORMAT;
            return 0;
        }

        // Da die Tabellendefinitionen fuer die Zutrittskontrolle eine
        // feste Struktur haben, wird aufgrund des 0 - Zeigers im Strukt
        // und des Flags fuer Auslesen, im Setup das Einfuegen vorgenommen.
        extBegin.offsetAccessControlEventChainsV1 = extBegin.offsetFreeScope;

        // Groesse der Zutrittskontrolle - Eingabeketten bereits im Eep?
        ulTmp = *(unsigned int*)&m_file.m_eep[P_ENTRANCE_EK_SIZE];
        ChangeOrderLong((unsigned char*)&ulTmp);
        // Liegt im moeglichen Bereich?
        if (!(ulTmp > EXTENDED_SIZE_SU) && !(ulTmp == 0))
        {
            ulSize = ulTmp;
        }
        else
        {			
            *piError = IV_ERROR_FILE_FORMAT;
            return 0;
        }

        // Zutritts - Eingabeketten - lesen
        if (!ReadMem(MB_ENTRANCE_EK, ulSize, &m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV1], piError))
        {
            return 0;
        }

        // Auf naechsten freien Speicher
        extBegin.offsetFreeScope += GetEntranceEKFlashSize();
    }
    else
    {
        // Auf naechsten freien Speicher
        extBegin.offsetFreeScope += GetMDEEKFlashSize_0400();
    }

    // Der Zeiger auf den naechsten freien Speicherplatz in der
    // erweiterten Struktur muss aufgrund der Abwaertskompatibilitaet
    // auf 0 gesetzt bleiben.
    extBegin.offsetFreeScope = 0;
    //*(TExtendedFlash*)m_file.m_flashExtended = extBegin;

    temp = 0; // Bedienstopp ruecksetzen
    if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
    {
        return 0;
    }

    res = 1;
    unsigned int sizeSetup, flashSizeSU2;
    // Setupdatei oeffnen
    FILE *pFile;
    if ((pFile = fopen(szFileName, "wb")) == NULL)
    {
        *piError = IV_ERROR_FILE_OPEN;
        return 0;
    }

    // EEPROM - schreiben
    if (fwrite(m_file.m_eep, sizeof(unsigned char), EEP_SIZE_SU, pFile) != EEP_SIZE_SU)
    {
        // Fehler beim schreiben in die Datei.
        *piError = IV_ERROR_FILE_WRITE;
        //_channel4.Trace("EEP konnte nicht geschrieben werden. ( %i )", ferror(pFile));
        res = 0;
        goto error_label;
    }

    // Dateikopf schreiben
    unsigned char mem[EEP_SIZE - EEP_SIZE_SU];
    memset(mem, 0xFF, sizeof(mem));
    if (fwrite(mem, sizeof(unsigned char), sizeof(mem), pFile) != sizeof(mem))
    {
        // Fehler beim schreiben in die Datei.
        *piError = IV_ERROR_FILE_WRITE;
        //_channel4.Trace("EEP konnte nicht geschrieben werden. ( %i )", ferror(pFile));
        res = 0;
        goto error_label;
    }

    // Groesse der Flashdaten fuer Setup aus EepMem entnehmen
    sizeSetup = *(unsigned int*)&m_file.m_eep[P_FLASH_SIZE_SU], flashSizeSU2 = 0;
    ChangeOrderLong((unsigned char*)&sizeSetup);
    // Groesse des zu schreibenden Flashbereichs ermitteln.
    if (sizeSetup > FLASH_SIZE_SU)
    {
        flashSizeSU2 = FLASH_SIZE_SU2;
    }

    if (res)
    {
        // Leere Filelist schreiben
        unsigned char buf[MAX_DATENSATZ_BESCHREIBUNG];
        memset(buf, 0, MAX_DATENSATZ_BESCHREIBUNG);

        // Flashbereich - schreiben
        if (fwrite(m_file.m_flash, sizeof(unsigned char), FLASH_SIZE_SU + flashSizeSU2, pFile) != (FLASH_SIZE_SU + flashSizeSU2) ||
            // Leere Filelist schreiben
            fwrite(buf, sizeof(unsigned char), MAX_DATENSATZ_BESCHREIBUNG, pFile) != MAX_DATENSATZ_BESCHREIBUNG ||
            // Konfiguration - schreiben
            fwrite(m_file.m_configMem, sizeof(unsigned char), CONFIG_SIZE_SU, pFile) != CONFIG_SIZE_SU ||
            // Erweiterter Flashbereich - schreiben
            fwrite(m_file.m_flashExtended, sizeof(unsigned char), EXTENDED_SIZE_SU, pFile) != EXTENDED_SIZE_SU /*||
            // Leere Filelist fuer Zutrittskontrolle (Version 1)
            fwrite(buf, sizeof(unsigned char), 4, pFile) != 4 ||
            // Leere Filelist fuer Zutrittskontrolle (Version 2)
            fwrite(buf, sizeof(unsigned char), MAX_DATENSATZ_BESCHREIBUNG, pFile) != MAX_DATENSATZ_BESCHREIBUNG*/)
        {
            // Fehler beim schreiben in die Datei.
            *piError = IV_ERROR_FILE_WRITE;
            //_channel4.Trace("Datenbereich konnte nicht geschrieben werden. ( %i )", ferror(pFile));
            res = 0;
            goto error_label;
        }
    }

error_label:
    if (pFile)
    {		
        // Datei schlieszen
        fclose(pFile);
    }

    return res;
}

int CChannel1_2::CheckDevice( int iPollRetry, int *piError )
{
    for (_devicePollRetry=1; _devicePollRetry <= iPollRetry; _devicePollRetry++)
    {
        int tmp;
        if (ReadMem(0ul, 2, (unsigned char*)&tmp, piError))
        {
            return 1;
        }

        Sleep(20);
    }

    return 0;
}

int CChannel1_2::Restart(int *piError)
{
    int tmp = 0;

    if (WriteMem(MB_RESTART, 1, (unsigned char*)&tmp, piError))
    {
        return 1;
    }

    return 0;
}

int CChannel1_2::GetSerialNum(int *piSerialNum, int *piError)
{
    unsigned char buf[10];

    // Seriennummer und Optionsbyte lesen
    if (!ReadMem(MB_SERIEN_MEMORY, 4, buf, piError))
    {
        return 0;
    }

    *piSerialNum = (int)(buf[2] << 8);
    *piSerialNum |= (int)(buf[3]);
    
    return 1;
}

// ListenBuffer in AE laden
int CChannel1_2::WriteEntranceListData(int iListNum, int *piError)
{
    // Alle Listen laden?
    if (iListNum == -1)
    {
        // Mindestens eine Liste geladen?
        int i;
        for (i=0; i < 4; i++)
        {
            if (m_ulEntranceListSize[i] != 0)
            {
                break;
            }
        }
        if (i == 4)
        {
            *piError = IV_ERROR_LIST_PROCESSING;
            return 0;
        }
    }
    else
    {
        if (m_ulEntranceListSize[iListNum] == 0)
        {
            *piError = IV_ERROR_LIST_PROCESSING ;
            return 0; // keine Listen vorhanden
        }
    }

    unsigned char temp = 1; // Bedienstopp setzen
    // Der Bedienstopp wird nur noch bei aelteren Versionen gesetzt.
    // Wenn der Befehl zum Bedienstopp aufheben gegeben wird,
    // dann werden die zuvor uebertragenen Daten an die entsprechende Position kopiert.
    switch (GetProtocolVersion2(GetDeviceIDRemote(), 1, piError))
    {
    case 0x0501:
    case 0x0502:
        break;

    default:
        if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
        {
            return 0;
        }
    }

    switch (iListNum)
    {
    case -1:
    case 0:
        if (m_ulEntranceListSize[0] != 0) 
        {
            // Flash schreiben 
            if (!WriteMem(MB_ENTRANCE_TABLE_0, m_ulCompressedListSize[0], m_pucCompressedListMem[0], piError, 32))
            {
                return 0;
            }
        }
        if (iListNum != -1) break;
    case 1:
        if (m_ulEntranceListSize[1] != 0)
        {
            // Flash schreiben 
            if (!WriteMem(MB_ENTRANCE_TABLE_1, m_ulCompressedListSize[1], m_pucCompressedListMem[1], piError, 32))
            {
                return 0;
            }
        }
        if (iListNum != -1) break;
    case 2:
        if (m_ulEntranceListSize[2] != 0)
        {
            // Flash schreiben 
            if (!WriteMem(MB_ENTRANCE_TABLE_2, m_ulCompressedListSize[2], m_pucCompressedListMem[2], piError, 32))
            {
                return 0;
            }
        }
        if (iListNum != -1) break;
    case 3:
        if (m_ulEntranceListSize[3] != 0)
        {
            // Flash schreiben 
            if (!WriteMem(MB_ENTRANCE_TABLE_3, m_ulCompressedListSize[3], m_pucCompressedListMem[3], piError, 32))
            {
                return 0;
            }
        }
    }

    temp = 0; // Bedienstopp ruecksetzen
    if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
    {
        return 0;
    }

    return 1;
}

int CChannel1_2::GetVersionFirmware(char *buf, int *piError)
{
    unsigned char lBuf[32];

    if (m_protocolVersion == 1) // Nur Protokollversion 1
    {
        switch (GetProtocolVersion2(GetDeviceIDRemote(), 1, piError))
        {
        case 0:
            return 0;

        case 0x0400: // AEIII+
        case 0x0401:
            *piError = IV_ERROR_VERSION_FIRMWARE;
            return 0;
        }
    }

    // Firmwareversion lesen
    if (!ReadMem(MB_VERSION, 32, lBuf, piError))
    {
        return 0;
    }
    memcpy(buf, lBuf, 32);

    return 1;
}

int CChannel1_2::WriteListData( Channel4Buffer *buffer, int *piError )
{
    int i;
    // Listengesamtgroesse berechnen und ggf. sortieren.
    unsigned int listSize = MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int);
    unsigned int listPointer[MAX_DATENSATZ_BESCHREIBUNG];
    for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++)
    {
        if (buffer->m_listSize[i])
        {			
            listPointer[i] = listSize + FLASH_LIST_START;
            ChangeOrderLong((unsigned char*)&listPointer[i]);

            listSize += buffer->m_listSize[i];
        }
        else
        {
            listPointer[i] = 0;
        }
    }

    if (listSize == (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int)))
    {
        *piError = IV_ERROR_LIST_PROCESSING;
        return 0; // keine Listen vorhanden
    }

    unsigned int listStart = FLASH_LIST_START;
    unsigned char temp = 1; // Bedienstopp setzen
    // Der Bedienstopp wird nur noch bei aelteren Versionen gesetzt.
    // Wenn der Befehl zum Bedienstopp aufheben gegeben wird,
    // dann werden die zuvor uebertragenen Daten an die entsprechende Position kopiert.
    switch (GetProtocolVersion2(GetDeviceIDRemote(), 1, piError))
    {
    case 0: // Kommunikationsfehler
        return 0;

    case 0x0501:
        // Listengroesse nur maximal 100kByte
        if (listSize >= FLASH_SIZE_LIST)
        {
            *piError = IV_ERROR_LIST_PROCESSING;
            return 0;
        }
        break;

    case 0x0502:
        listStart = FLASH_LIST_START_NEW;

        // Listengroesse nur maximal 244kByte
        if (listSize >= FLASH_SIZE_LIST_NEW)
        {
            *piError = IV_ERROR_LIST_PROCESSING;
            return 0;
        }
        break;

    default:
        // Listengroesse nur maximal 100kByte
        if (listSize >= FLASH_SIZE_LIST)
        {
            *piError = IV_ERROR_LIST_PROCESSING;
            return 0;
        }
        if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
        {
            return 0;
        }
    }

    // Flash schreiben 
    if (!WriteMem(listStart, (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int)), (unsigned char*)listPointer, piError))
    {
        return 0;
    }

    unsigned int tmpSize = 0;
    for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++)
    {
        if (buffer->m_listSize[i])
        {
            // Datensatzanzahl drehen
            ChangeOrderLong((unsigned char*)&buffer->m_pucListMem[i][0]);

            // Flash schreiben 
            if (!WriteMem(listStart + (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int)) + tmpSize, buffer->m_listSize[i], buffer->m_pucListMem[i], piError))
            {
                // Datensatzanzahl drehen
                ChangeOrderLong((unsigned char*)&buffer->m_pucListMem[i][0]);
                return 0;
            }

            // Datensatzanzahl drehen
            ChangeOrderLong((unsigned char*)&buffer->m_pucListMem[i][0]);
            tmpSize += buffer->m_listSize[i];
        }
    }

    temp = 0; // Bedienstopp ruecksetzen
    if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
    {
        return 0;
    }	

    return 1;
}

// Prüfen ob die Displaydaten in Ordnung sind.
int CChannel1_2::CheckFileDisplay(FileDisplay *pFileDisplay, int *piError)
{
    unsigned int ulFontPtr, ulGraficPtr;
    int i, iFontCount, iGraficCount;
    FileDisplay fileDisplay;

    // Fehlerobjekt loeschen
    *piError = IV_ERROR_SUCCESS;

    // EEPROM - Zeiger auf Zeichensatz lesen 
    if (!ReadMem(MB_START_EEP + 0x0F78, sizeof(unsigned int), (unsigned char*)&ulFontPtr, piError))
    {
        return 0;
    }
    ChangeOrderLong((unsigned char*)&ulFontPtr);
    if (ulFontPtr == 0xFFFFFFFF)
    { 
        // Zeiger nicht vorhanden
        *piError = IV_ERROR_SUCCESS;
        return 0;
    }

    // EEPROM - Zeiger auf Grafik lesen 
    if (!ReadMem(MB_START_EEP + 0x0F7C, sizeof(unsigned int), (unsigned char*)&ulGraficPtr, piError))
    {
        return 0;
    }
    ChangeOrderLong((unsigned char*)&ulGraficPtr);
    if (ulGraficPtr == 0xFFFFFFFF)
    { 
        // Zeiger nicht vorhanden
        *piError = IV_ERROR_SUCCESS;
        return 0;
    }

    // Hauptstruktur aus Flash lesen um Anzahl zu ermitteln 
    if (!ReadMem(0x000B0200 + ulFontPtr, sizeof(FONTNODE), fileDisplay.m_memFont, piError))
    {
        return 0;
    }
    // Anzahl Zeichensaetze
    iFontCount = fileDisplay.m_pFontNode->cnt;
    if ((iFontCount > MAX_FONT) || (iFontCount != pFileDisplay->m_pFontNode->cnt))
    {
        *piError = IV_ERROR_SUCCESS;
        return 0;
    }

    // Hauptstruktur aus Flash lesen um Anzahl zu ermitteln 
    if (!ReadMem(0x000B0200 + ulGraficPtr, sizeof(GRAFICNODE), fileDisplay.m_memGrafic, piError))
    {
        return 0;
    }
    // Anzahl Grafiken
    iGraficCount = fileDisplay.m_pGraficNode->cnt;
    if ((iGraficCount > MAX_GRAFIC) || (iGraficCount != pFileDisplay->m_pGraficNode->cnt))
    {
        *piError = IV_ERROR_SUCCESS;
        return 0;
    }
    // Klassenstrukturen initialisieren
    fileDisplay.Default(iFontCount, iGraficCount);

    // Strukturen aus Flash lesen 
    if (!ReadMem(0x000B0200 + ulFontPtr, sizeof(FONTNODE) + iFontCount * sizeof(unsigned int) + iFontCount * sizeof(FONT), fileDisplay.m_memFont, piError))
    {
        return 0;
    }
    // CRC von Zeichensaetzen ueberpruefen
    for (i=0; i < iFontCount; i++)
    {
        if (fileDisplay.m_pFont[i]->crc != pFileDisplay->m_pFont[i]->crc)
        {
            *piError = IV_ERROR_SUCCESS;
            return 0;
        }
    }

    // Strukturen aus Flash lesen 
    if (!ReadMem(0x000B0200 + ulGraficPtr, sizeof(GRAFICNODE) + iGraficCount * sizeof(unsigned int) + iGraficCount * sizeof(GRAFIC), fileDisplay.m_memGrafic, piError))
    {
        return 0;
    }
    // CRC von Grafiken ueberpruefen
    for (i=0; i < iGraficCount; i++)
    {
        if (fileDisplay.m_pGrafic[i]->crc != pFileDisplay->m_pGrafic[i]->crc)
        {
            *piError = IV_ERROR_SUCCESS;
            return 0;
        }
    }

    // Zeichensatz- und Grafikdaten vorhanden und identisch
    return 1;
}

int CChannel1_2::WriteFileDevice( char *szFileName, int *piError )
{
    // Hardware- und Softwareinfo der ausgewaehlten Datei.
    StructHardwareInfo hwInfo;

    int iRes = IV_ERROR_SUCCESS;

    // Puffer fuer die Hardwareinfo in Textform
    memset(_info, 0, sizeof(_info));
    // Strukturen loeschen.
    memset(&hwInfo, 0, sizeof(StructHardwareInfo));


    //#########################################################################
    // Pruefen ob gewaehlte Datei in das Geraet eingespielt werden kann.
    if ( m_protocolVersion == 1) // AEIII+ nicht unterstuetzt
    {
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;
    }

    // Dateityp ermitteln
    struct stat st;
    if (stat(szFileName, &st) == -1) 
    {	
        *piError = IV_ERROR_FILE_OPEN;
        return 0;
    }

    // Angabe ist eine Datei
    if (S_ISREG(st.st_mode)) 
    {	
        // Nur ".hex" gueltig.
        size_t pos = strlen(szFileName) - 4;
#ifdef _MSC_VER
        if (szFileName[pos] != '.' || 
            ((!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
            !(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
            !(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X')) &&
            (!(szFileName[pos + 1] == 'd' || szFileName[pos + 1] == 'D') ||
            !(szFileName[pos + 2] == 'f' || szFileName[pos + 2] == 'F') ||
            !(szFileName[pos + 3] == 'z' || szFileName[pos + 3] == 'Z')) &&
            (!(szFileName[pos + 1] == 'z' || szFileName[pos + 1] == 'Z') ||
            !(szFileName[pos + 2] == 'i' || szFileName[pos + 2] == 'I') ||
            !(szFileName[pos + 3] == 'p' || szFileName[pos + 3] == 'P'))) ) 
        {	
            *piError = IV_ERROR_FILE_EXTENSION;
            return 0;
        }
#else 
        if (!S_ISREG(st.st_mode) || szFileName[pos] != '.' || 
            !(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
            !(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
            !(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X') ) 
        {
            *piError = IV_ERROR_FILE_EXTENSION;
            return 0;
        }
#endif	
        // Handelt es sich um eine Hexdatei
        if (szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') 
        {
            // Dateidaten laden
            m_fileDevice.reset();
            if (!m_fileDevice.load(szFileName)) 
            {
                if (m_fileDevice.errorState() == FileIntelHex::Version || m_fileDevice.errorState() == FileIntelHex::DeviceUnknown) 
                {
                    *piError = IV_ERROR_VERSION_DEVICEFILE;
                } 
                else if (m_fileDevice.errorState() == FileIntelHex::FileOpen) 
                {
                    *piError = IV_ERROR_FILE_OPEN;
                } 
                else 
                {
                    *piError = IV_ERROR_DEVICEFILE_LOAD;
                }

                m_fileDevice.reset();
                return 0;
            }

            // Ist die Softwareinformation in der Geraetedatei verfuegbar?
            if (m_fileDevice.GetHeaderSoftwareInfo().size) 
            {	
                // Auslesen der Hardwareinfo
                if (!GetHardwareInfo(GetDeviceIDRemote(), hwInfo, piError)) 
                {
                    return 0;
                } // Hardwareinfo

                // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
                if (0 == m_fileDevice.CompareInfos(hwInfo, m_fileDevice.GetHeaderSoftwareInfo())) 
                {	
                    // Hardwareinfo in Text aufloesen
                    //_channel4.getInfos(hwInfo, _info);
                    m_fileDevice.setErrorTextExtended( m_fileDevice.errorText() );
                    *piError = IV_ERROR_DEVICEFILE_MATCH;
                    return 0;
                }
            } 
            else 
            {
                std::string version;
                // Die Geraetedatei muss vom Typ "PZE-MasterIV" sein, oder diesen unterstuetzen.
                if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_PZE) 
                {
                    // Die gewaehlte Geraetedatei ist NICHT fuer das PZE-MasterIV.
                    *piError = IV_ERROR_DEVICEFILE_WRONG;
                    return 0;
                }

                // Da Protokollversion 2 muss es sich um eine Boardversion < 3 handeln.
                // Es muss sich um eine 04.00.xx oder 04.01.xx Firmwareversion handeln.
                if (m_fileDevice.GetHeaderVersionFirmware(version)) 
                {
                    if (version.compare( 0, 6, "04.00." ) != 0 && version.compare( 0, 6, "04.01." ) != 0)
                    {
                        *piError = IV_ERROR_DEVICEFILE_MATCH;
                        return 0;
                    }
                }
            }
        } 
        else 
        { // Es handelt sich um eine Zipdatei

            // Auslesen der Hardwareinfo
            if (!GetHardwareInfo(GetDeviceIDRemote(), hwInfo, piError)) 
            {	
                return 0;
            }

            // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
            if (0 == m_fileDevice.SearchInfosZip(szFileName, hwInfo)) 
            {	
                // Hardwareinfo in Text aufloesen
                //_channel4.getInfos(hwInfo, _info);
                if (m_fileDevice.errorState() == FileIntelHex::Version || m_fileDevice.errorState() == FileIntelHex::DeviceUnknown) 
                {
                    *piError = IV_ERROR_VERSION_DEVICEFILE;
                } 
                else if (m_fileDevice.errorState() == FileIntelHex::NoFileMatch) 
                {
                    *piError = IV_ERROR_DEVICEFILE_MATCH;
                } 
                else 
                {
                    *piError = IV_ERROR_DEVICEFILE_LOAD;
                }

                return 0;
            }
        }	
    } 
    else if (S_ISDIR(st.st_mode)) 
    {
#ifdef _MSC_VER
        // Auswahl aus Verzeichnis nur unter Linux gueltig!
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return 0;
#else
        // Auslesen der Hardwareinfo
        if (!GetHardwareInfo(GetDeviceIDRemote(), hwInfo, piError)) 
        {
            return FALSE;
        }

        // Pruefen ob eine passende Datei in die vorliegende Hardware eingespielt werden kann.
        if (0 == m_fileDevice.SearchInfosDir(szFileName, hwInfo)) 
        {	
            // Hardwareinfo in Text aufloesen
            //((Channel4*)&GetChannelBase())->getInfos(hwInfo, _info);
            if (m_fileDevice.errorState() == FileIntelHex::Version || m_fileDevice.errorState() == FileIntelHex::DeviceUnknown)
            {
                *piError = IV_ERROR_VERSION_DEVICEFILE;
            } 
            else if (m_fileDevice.errorState() == FileIntelHex::NoFileMatch)
            {
                *piError = IV_ERROR_DEVICEFILE_MATCH;
            } 
            else 
            {
                *piError = IV_ERROR_DEVICEFILE_LOAD;
            }

            return 0;
        }
#endif
    } 
    else 
    {	
        *piError = IV_ERROR_FILE_OPEN;
        return 0;
    }	

            // Geraetemodus ermitteln
        int modus = GetModus(piError);
        if (!modus) 
        {
            return 0;
        }

        // Applikationsmodus? 
        if (modus == 2) 
        {
            // Abfrage ob Daten vorhanden
            if ((iRes = CheckData(piError)) == -1) 
            {
                return 0;
            } 
            else if (iRes) 
            {
                *piError = IV_ERROR_DATA_AVAILABLE;
                return 0;
            }
        }

        // Geraetetexte ggf als Zeichensatz anfuegen
        if (m_fileDevice.GetFileDeviceText()->m_sizeData) 
        {
            // Zeichensatz- und Grafikdaten
            //CDFFileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();

            FileDeviceText *fileDeviceText = m_fileDevice.GetFileDeviceText();
            unsigned char *buf = fileDeviceText->m_memData;
            unsigned int size = (unsigned int)fileDeviceText->m_sizeData;

            // Texte als Zeichensatz mit uebertragen.
            int err;
            if (!(err = m_fileDevice.GetFileDisplay()->AppendSegmentText(buf, size))) 
            {
                if (FileIntelHex::AllocMem == m_fileDevice.GetFileDisplay()->errorState()) 
                {
                    *piError = IV_ERROR_ALLOC_MEM;
                    return 0;
                }

                *piError = IV_ERROR_FILE_PROCESSING;
                return 0;
            }

            // Ruecksetzen des Flags fuer "gueltige Default-Textdaten vorhanden"
            unsigned char v = 0;
            if (!WriteMem(MB_START_EEP + 0x0ECB, sizeof(unsigned char), (unsigned char*)&v, piError)) 
            {
                return 0;
            }
        }

        unsigned char loadDisplayData = 1;
        unsigned int sizeData = m_fileDevice.GetSizeData();
        unsigned int tmp = ~0;
        unsigned int address;
        unsigned int firstAddress = m_fileDevice.firmwareFirstAddress();
        unsigned int lastAddress  = m_fileDevice.firmwareLastAddress();
        unsigned int sizeImage   = lastAddress - firstAddress + 1;
		std::vector<unsigned char> firmwareData = m_fileDevice.firmwareContent();

        unsigned int crc = 0;

        // EEPROM - Setupzeiger -pKonf- loeschen
        //@@: Uebergangsbug wegen Speicherueberlappung, von Firmware und Setup, beseitigen
        if (!WriteMem(MB_START_EEP + 4, sizeof(unsigned int), (unsigned char*)&tmp, piError)) 
        {
            return 0;
        }

        /* Zeichensatz und Displaydaten immer uebertragen
        // Pruefen ob die Zeichensatz und Grafikdaten vorhanden
        if (!checkFileDisplay(m_fileDevice.GetFileDisplay(), piError)) 
        {
        if (*piError != IV_ERROR_SUCCESS) 
        {
        // Fehler waehrend Kommunikation, abbruch
        return 0;
        }

        // Keine Displaydaten oder ungleich!

        } 
        else 
        {
        // Anzeigedaten vorhanden und Identisch, also nicht uebertragen
        sizeData -= m_fileDevice.GetFileDisplay()->GetSizeData();
        loadDisplayData = 0;
        }*/

        // EEPROM - Bootload Beginn
        if (!WriteMem(MB_BOOTLOAD, sizeof(unsigned int), (unsigned char*)&sizeData, piError)) 
        {
            return 0;
        }

        tmp = FLASH_IMAGE_BASE + firstAddress; // Basisadresse im Flash
        // EEPROM - schreiben: Startadresse des Image im externen Flash
        if (!WriteMem(MB_START_EEP + EEPROM_IMAGE_SOURCE, sizeof(unsigned int), (unsigned char*)&tmp, piError)) 
        {
            return 0;
        }

        // EEPROM - schreiben: Zieladresse des Image
        if (!WriteMem(MB_START_EEP + EEPROM_IMAGE_DEST, sizeof(unsigned int), (unsigned char*)&firstAddress, piError)) 
        {
            return 0;
        }

        // EEPROM - schreiben: Laenge des Images
        if (!WriteMem(MB_START_EEP + EEPROM_IMAGE_LEN, sizeof(unsigned int), (unsigned char*)&sizeImage, piError)) 
        {
            return 0;
        }

        // CRC berechnen
		crc = GetCRC16( firmwareData.data(), firmwareData.size() );

        // EEPROM - schreiben: CRC des Images
        if (!WriteMem(MB_START_EEP + EEPROM_IMAGE_CRC, sizeof(unsigned int), (unsigned char*)&crc, piError)) 
        {
            return 0;
        }

        // Imagedaten
        for (address=firstAddress; address < (lastAddress + 1); address += MAX_MB_PAKET_DATA) 
        {
            // FLASH - schreiben
            if (!WriteMem(FLASH_IMAGE_BASE + address, 
                (((lastAddress + 1) - address) > MAX_MB_PAKET_DATA) ? MAX_MB_PAKET_DATA : ((lastAddress + 1) - address), 
                &firmwareData[address - firstAddress], piError)) 
            {
                return 0;
            }
        } 

        // Laden der Zeichensatz- und Grafikdaten?
        if (loadDisplayData) 
        {
            // Zeichensatz- und Grafikdaten
            FileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();

            // Zeichensatzdaten schreiben
            for (address=0; address < pFileDisplay->m_sizeFont; address += MAX_MB_PAKET_DATA) 
            {
                // Flash schreiben 
                if (!WriteMem(0x000B0200 + address, 
                    ((address + MAX_MB_PAKET_DATA) < pFileDisplay->m_sizeFont) ? MAX_MB_PAKET_DATA : pFileDisplay->m_sizeFont - address, 
                    &pFileDisplay->m_memFont[address], piError)) 
                {
                    return 0;
                }
            }

            // Grafikdaten schreiben
            for (address=0; address < pFileDisplay->m_sizeGrafic; address += MAX_MB_PAKET_DATA) 
            {
                if (!WriteMem(0x000B0200 + pFileDisplay->m_sizeFont + address, 
                    ((address + MAX_MB_PAKET_DATA) < pFileDisplay->m_sizeGrafic) ? MAX_MB_PAKET_DATA : pFileDisplay->m_sizeGrafic - address,
                    &pFileDisplay->m_memGrafic[address], piError)) 
                {
                    return 0;
                }
            }

            tmp = 0;
            ChangeOrderLong((unsigned char*)&tmp);
            // EEPROM - Zeiger auf Font setzen 
            if (!WriteMem(MB_START_EEP + 0x0F78, sizeof(unsigned int), (unsigned char*)&tmp, piError)) 
            {
                return 0;
            }

            tmp = pFileDisplay->m_sizeFont;
            ChangeOrderLong((unsigned char*)&tmp);
            // EEPROM - Zeiger auf Grafik setzen 
            if (!WriteMem(MB_START_EEP + 0x0F7C, sizeof(unsigned int), (unsigned char*)&tmp, piError)) 
            {
                return 0;
            }
        }

        // Schreiben der Geraetetexte
        // Noch einbauen!! 

        sizeData = 0;
        // EEPROM - Bootload Ende
        if (!WriteMem(MB_BOOTLOAD, sizeof(unsigned int), (unsigned char*)&sizeData, piError)) 
        {
            return 0;
        }

        return 1;
}

int CChannel1_2::GetVersionBootloader(char *version, int *piError)
{
    unsigned int ver;

    // Fehlerobjekt loeschen
    int error = IV_ERROR_SUCCESS;

    // EEPROM - Bootversion lesen 
    if (!ReadMem(MB_START_EEP + EEPROM_BOOT_VERSION, sizeof(unsigned int), (unsigned char*)&ver, &error)) {

        *piError = IV_ERROR_UNREACHABLE;
        return 0;
    }

    sprintf(version, "%02X.%02X.%02X.%02X", ((unsigned char*)&ver)[3], ((unsigned char*)&ver)[2], ((unsigned char*)&ver)[1], ((unsigned char*)&ver)[0]);

    return 1;
}

int CChannel1_2::WriteMemToFile(int type, int address, int size, char *fileName)
{
    StructMemFileHeader header;
    int res = IV_ERROR_SUCCESS, chunk;
    char buf[0x400];
    int baseAddress = 0, offsAddress = 0;

    FILE *fp = fopen(fileName, "wb");
    if (fp == NULL)
    {
        return -IV_ERROR_FILE_OPEN;
    }
    // Dateiheader aufsetzen
    memset(&header, 0, sizeof(header));
    header.structSize = sizeof(header);
    header.type = type;
    header.baseAddress = baseAddress;
    header.memAddress = address;
    header.memSize = size;
    switch (type)
    {
    case 0:
        baseAddress = MB_ACCESS_SRAM;
        break;

    case 1:
        baseAddress = MB_ACCESS_EEPROM;
        break;

    case 2:
        baseAddress = MB_ACCESS_FLASH;
        break;

    case 3:
        baseAddress = MB_ACCESS_PROGRAM;
        break;

    case 4:
        baseAddress = MB_ACCESS_FLASH_PAGE;
        break;

    case 5:
        baseAddress = MB_ACCESS_MMC;
        break;
    }

    // Basisaddress nachtragen
    header.baseAddress = baseAddress;
    // Dateiheader schreiben
    if (fwrite(&header, sizeof(header), 1, fp) != 1)
    {	
        res = -IV_ERROR_FAIL;
        goto label_out;
    }

    for ( ;; )
    {
        // Naechsten Block berechnen
        chunk = ((size - offsAddress) < (int)sizeof(buf)) ? size - offsAddress : sizeof(buf);
        if (chunk == 0) break; // Keine weiteren Daten

        if (!ReadMem(baseAddress + address + offsAddress, chunk, (unsigned char*)buf, &res))
        {
            res = -IV_ERROR_FAIL;
            goto label_out;
        }

        if (fwrite(buf, chunk, 1, fp) != 1)
        {
            res = -IV_ERROR_FAIL;
            goto label_out;
        }

        offsAddress += chunk;
    }

label_out:
    if (fp) fclose(fp);
    return res;
}

int CChannel1_2::WriteFileToMem(char *fileName)
{
    StructMemFileHeader header;
    int res = IV_ERROR_SUCCESS, chunk = 0;
    int offsAddress = 0;
    char buf[0x400];

    FILE *fp = fopen(fileName, "rb");
    if (fp == NULL)
    {
        return -IV_ERROR_FILE_OPEN;
    }

    // Dateiheader lesen
    if (fread(&header, sizeof(header), 1, fp) != sizeof(header))
    {	
        res = -IV_ERROR_FAIL;
        goto label_out;
    }

    offsAddress = header.baseAddress;
    for ( ;; )
    {
        // Naechsten Block berechnen
        chunk = ((header.memSize - offsAddress) < (int)sizeof(buf)) ? header.memSize - offsAddress : sizeof(buf);

        if (fread(buf, chunk, 1, fp) != (unsigned)chunk)
        {
            res = -IV_ERROR_FAIL;
            goto label_out;
        }

        if (!WriteMem(header.baseAddress + header.memAddress + offsAddress, chunk, (unsigned char*)buf, &res))
        {
            goto label_out;
        }

        offsAddress += chunk;
    }

label_out:
    if (fp) fclose(fp);
    return res;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int CChannel1_2::CheckData(int *piError)
{
    unsigned char ucIsData = FALSE;

    // Abfrage ob Daten vorhanden
    if (!ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError))
    {			
        return -1;
    }

    return ucIsData;
}

