/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#ifndef CHANNEL4_MODULE_HARDWARE_H
#define CHANNEL4_MODULE_HARDWARE_H

// Basisklasse einbinden.
#include "Channel4ProtocolCrypt.h"
// Datenklasse
#include "../Common/hardwareinfo.h"
#include <stdlib.h>

/** In diesem Module sind alle Hardwarespezifischen Funktionen enthalten.
*
* Die Aufteilung basiert Hauptsaechlich auf Typen, Komponenten und Optionen.
*
*/
#define IV_MODULE_HARDWARE                       0x0000

// Setzen/Ermitteln des Geraetetyps
#define IV_FUNCTION_SET_TYPE_DEVICE              0x0001
#define IV_FUNCTION_GET_TYPE_DEVICE              0x0002

#define IV_VALUE_TYPE_DEVICE_COUNT                         23 // Gesamtzahl der Geraetetypen
#define IV_VALUE_TYPE_DEVICE_ALL                          (~0)// Alle Typen werden unterstuetzt.
#define IV_VALUE_TYPE_DEVICE_PZE                            0 // PZE-MasterIV
#define IV_VALUE_TYPE_DEVICE_ZK                             1 // ZK-MasterIV
#define IV_VALUE_TYPE_DEVICE_BDE                            2 // BDE-MasterIV
#define IV_VALUE_TYPE_DEVICE_MOBIL                          3 // Mobil-MasterIV
#define IV_VALUE_TYPE_DEVICE_MDE                            4 // MDE-BoxIV
#define IV_VALUE_TYPE_DEVICE_AE                             5 // AE-MasterIV
#define IV_VALUE_TYPE_DEVICE_TIMEBOY                        6 // TimeboyIV
#define IV_VALUE_TYPE_DEVICE_EXKLUSIV                       7 // Exklusiv-Line IV
#define IV_VALUE_TYPE_DEVICE_FLEX                           8 // Flex-MasterIV
#define IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE              9 // Timeboy-Mobil-PZE
#define IV_VALUE_TYPE_DEVICE_TERMINAL_2_8                  10 // EVO 2.8
#define IV_VALUE_TYPE_DEVICE_TERMINAL_4_3                  11 // EVO 4.3
#define IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7       12 // IPC-Extension
#define IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_10        13 // IPC-Extension für Vario 10 / Vario 12 / Vario 15
#define IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_Q7              14 // IPC-Extension für Q7
#define IV_VALUE_TYPE_DEVICE_IO_BOX                        15 // IO-Box
#define IV_VALUE_TYPE_DEVICE_ZK_BOX                        16 // ZK-Box
#define IV_VALUE_TYPE_DEVICE_MOBIL_BOX                     17 // Mobil-Box
#define IV_VALUE_TYPE_DEVICE_TIMEBOY_TBDOCK                18 // Timeboy-Docking
#define IV_VALUE_TYPE_DEVICE_MOBIL_TBDOCK                  19 // Mobil-Docking
#define IV_VALUE_TYPE_DEVICE_ZK_KNOTEN                     20 // ZK-Knoten
#define IV_VALUE_TYPE_DEVICE_ZK_TERMINAL                   21 // ZK-Terminal
#define IV_VALUE_TYPE_DEVICE_SAMS_OCL_READER               22 // SAMS-OCL-Reader

// Typen in Textform
#define IV_VALUE_TYPE_DEVICE_NAME                          "MasterIV"
#define IV_VALUE_TYPE_DEVICE_PZE_NAME                      "PZE-MasterIV"
#define IV_VALUE_TYPE_DEVICE_ZK_NAME                       "ZK-MasterIV"
#define IV_VALUE_TYPE_DEVICE_BDE_NAME                      "BDE-MasterIV"
#define IV_VALUE_TYPE_DEVICE_MOBIL_NAME                    "Mobil-MasterIV"
#define IV_VALUE_TYPE_DEVICE_MDE_NAME                      "MDE-BoxIV"
#define IV_VALUE_TYPE_DEVICE_AE_NAME                       "AE-MasterIV"
#define IV_VALUE_TYPE_DEVICE_TIMEBOY_NAME                  "TimeboyIV"
#define IV_VALUE_TYPE_DEVICE_EXKLUSIV_NAME                 "Exklusiv-Line IV"
#define IV_VALUE_TYPE_DEVICE_FLEX_NAME                     "Flex-MasterIV"
#define IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE_NAME        "TimeboyIV Mobil-PZE"
#define IV_VALUE_TYPE_DEVICE_TERMINAL_2_8_NAME             "EVO 2.8"
#define IV_VALUE_TYPE_DEVICE_TERMINAL_4_3_NAME             "EVO 4.3"
#define IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_NAME            "IPC Embedded-System"
#define IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7_NAME  "IPC-Extension A"
#define IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_10_NAME   "IPC-Extension B"
#define IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_Q7_NAME         "IPC-Extension C"
#define IV_VALUE_TYPE_DEVICE_IO_BOX_NAME                   "IO-Box V4"
#define IV_VALUE_TYPE_DEVICE_ZK_BOX_NAME                   "ZK-Box V4"
#define IV_VALUE_TYPE_DEVICE_MOBIL_BOX_NAME                "Mobil-Box V4"
#define IV_VALUE_TYPE_DEVICE_TIMEBOY_TBDOCK_NAME           "Dockingstation V2"
#define IV_VALUE_TYPE_DEVICE_MOBIL_TBDOCK_NAME             "Fahrzeugdatenlogger V2"
#define IV_VALUE_TYPE_DEVICE_ZK_KNOTEN_NAME                "ZK-Knoten"
#define IV_VALUE_TYPE_DEVICE_ZK_TERMINAL_NAME              "ZK-Terminal"
#define IV_VALUE_TYPE_DEVICE_SAMS_OCL_READER_NAME          "SAMS-OCL-Reader"

// Setzen/Ermitteln des Boardtyp (Redesignversion z.B.: 0x13) (in Nibbles)
#define IV_FUNCTION_SET_TYPE_BOARD               0x0003
#define IV_FUNCTION_GET_TYPE_BOARD               0x0004

#define IV_VALUE_TYPE_BOARD_1_0                  0x10
#define IV_VALUE_TYPE_BOARD_1_1                  0x11
#define IV_VALUE_TYPE_BOARD_1_2                  0x12
#define IV_VALUE_TYPE_BOARD_1_3                  0x13
#define IV_VALUE_TYPE_BOARD_1_4                  0x14
#define IV_VALUE_TYPE_BOARD_2_0                  0x20
#define IV_VALUE_TYPE_BOARD_2_1                  0x21
#define IV_VALUE_TYPE_BOARD_3_0                  0x30
#define IV_VALUE_TYPE_BOARD_3_4                  0x34
#define IV_VALUE_TYPE_BOARD_4_0                  0x40
#define IV_VALUE_TYPE_BOARD_4_1                  0x41
#define IV_VALUE_TYPE_BOARD_4_2                  0x42
#define IV_VALUE_TYPE_BOARD_4_3                  0x43
#define IV_VALUE_TYPE_BOARD_2_3                  0x23
#define IV_VALUE_TYPE_BOARD_2_4                  0x24
#define IV_VALUE_TYPE_BOARD_3_1                  0x31
#define IV_VALUE_TYPE_BOARD_4_4                  0x44
#define IV_VALUE_TYPE_BOARD_1_5                  0x15

// Setzen/Ermitteln der Boardkomponente: RS485 fuer Kommunikation mit PC
#define IV_FUNCTION_SET_COMPONENT_RS485_0        0x0005
#define IV_FUNCTION_GET_COMPONENT_RS485_0        0x0006

// Setzen/Ermitteln der Boardkomponente: TCP/IP
#define IV_FUNCTION_SET_TYPE_TCPIP               0x0007
#define IV_FUNCTION_GET_TYPE_TCPIP               0x0008

#define IV_VALUE_TYPE_TCPIP_NONE                 0 // Kein TCP/IP
#define IV_VALUE_TYPE_TCPIP_XPORT                1 // Lantronix XPort
#define IV_VALUE_TYPE_TCPIP_MATCHPORT            2 // Lantronix MatchPort
#define IV_VALUE_TYPE_TCPIP_RS9110_N_11_24       3 // WLAN Redpine RS9110-N-11-24 Wird über typeModemMC gehandhabt.

// Setzen/Ermitteln der Boardkomponente: Modem MC35i/MC55
#define IV_FUNCTION_SET_TYPE_MODEM_MC            0x0009
#define IV_FUNCTION_GET_TYPE_MODEM_MC            0x000A

#define IV_VALUE_TYPE_MODEM_MC_NONE              0 // Kein GSM/GPRS
#define IV_VALUE_TYPE_MODEM_MC_39                1 // Siemens MC35i/MC39
#define IV_VALUE_TYPE_MODEM_MC_55                2 // Siemens MC55
#define IV_VALUE_TYPE_MODEM_RS9110_N_11_24       3 // WLAN Redpine RS9110-N-11-24

// Setzen/Ermitteln der Boardkomponente: RS485 fuer Zutrittskontrolle
#define IV_FUNCTION_SET_COMPONENT_RS485_1        0x000B
#define IV_FUNCTION_GET_COMPONENT_RS485_1        0x000C

// Setzen/Ermitteln des Transpondertyps
#define IV_FUNCTION_SET_TYPE_TRANSPONDER         0x000D
#define IV_FUNCTION_GET_TYPE_TRANSPONDER         0x000E

#define IV_VALUE_TYPE_TRANSPONDER_NONE                 0 // Kein Transponderleser
#define IV_VALUE_TYPE_TRANSPONDER_TSR30                1 // GIS TSR30
#define IV_VALUE_TYPE_TRANSPONDER_TSR32                2 // GIS TSR32
#define IV_VALUE_TYPE_TRANSPONDER_PRIMO100             3 // PHG Primo 100 (Legic Prime)
#define IV_VALUE_TYPE_TRANSPONDER_MIFARE               4 // Cavitec Mifare
#define IV_VALUE_TYPE_TRANSPONDER_IBUTTON_EXTERN       5 // Dallas iButton (Als separates Modul angebunden)
#define IV_VALUE_TYPE_TRANSPONDER_SMARTRELAIS          6 // SimonsVoss SmartRelais
#define IV_VALUE_TYPE_TRANSPONDER_NEDAP                7 // Nedap
#define IV_VALUE_TYPE_TRANSPONDER_IBUTTON_INTERN       8 // Dallas iButton (Fest auf Platine bestueckt)
#define IV_VALUE_TYPE_TRANSPONDER_HID_PROXKEY          9 // HID Proxy 4065
#define IV_VALUE_TYPE_TRANSPONDER_TWN3_MULTI_ISO      10 // Mifare Desfire (ISO14443 / ISO15693)
#define IV_VALUE_TYPE_TRANSPONDER_PRIMO130            11 // PHG Primo 130 (Legic Prime/Advant)
#define IV_VALUE_TYPE_TRANSPONDER_TSR21               12 // GIS TSR21
#define IV_VALUE_TYPE_TRANSPONDER_HID_ICLASS          13 // HID iClass OEM50
#define IV_VALUE_TYPE_TRANSPONDER_TB_MIFARE           14 // Mifare Lesemodul fuer TimeboyIV
#define IV_VALUE_TYPE_TRANSPONDER_TB_LEGIC            15 // Legic Lesemodul fuer TimeboyIV
#define IV_VALUE_TYPE_TRANSPONDER_SMARTRELAIS_G2      16 // SimonsVoss SmartRelais_G2
#define IV_VALUE_TYPE_TRANSPONDER_TWN3_MIFARE_NFC     17 // TWN3 Mifare NFC (ISO14443)
#define IV_VALUE_TYPE_TRANSPONDER_IDTRONIC_UHF        18 // ID Tronic UHF (long range)

// Setzen/Ermitteln des Displaytyp
#define IV_FUNCTION_SET_TYPE_DISPLAY             0x000F
#define IV_FUNCTION_GET_TYPE_DISPLAY             0x0010

#define IV_VALUE_TYPE_DISPLAY_NONE               0 // Kein Display
#define IV_VALUE_TYPE_DISPLAY_SMALL              1 // Lehner Dabitros (320 x 240) 4 Zoll Monochrom
#define IV_VALUE_TYPE_DISPLAY_BIG                2 // Lehner Dabitros (320 x 240) 6 Zoll Monochrom
#define IV_VALUE_TYPE_DISPLAY_SMALLER            3 // 1,9 Zoll Color
#define IV_VALUE_TYPE_DISPLAY_TIMEBOY_2ZOLL      4 // 2 Zoll Color Ilitek für TimeboyIV

// Setzen/Ermitteln des Tastaturtyps
#define IV_FUNCTION_SET_TYPE_KEYPAD              0x0011
#define IV_FUNCTION_GET_TYPE_KEYPAD              0x0012

#define IV_VALUE_TYPE_KEYPAD_NONE                0 // Keine Tastatur
#define IV_VALUE_TYPE_KEYPAD_9                   1 // 9 Tasten Tastatur (PZE-MasterIV usw.)
#define IV_VALUE_TYPE_KEYPAD_58                  2 // 56+2 Tasten Tastatur (BDE-MasterIV usw.)
#define IV_VALUE_TYPE_KEYPAD_29                  3 // 28+1 Tasten Tastatur (AE-MasterIV usw.)
#define IV_VALUE_TYPE_KEYPAD_7                   4 // 7 Tasten Tastatur (PZE-MasterIV Basic)
#define IV_VALUE_TYPE_KEYPAD_25                  5 // 25 Tasten Tastatur (TimeboyIV usw.)
#define IV_VALUE_TYPE_KEYPAD_20                  6 // 20 Tasten Tastatur (Flex-MasterIV usw.)
#define IV_VALUE_TYPE_KEYPAD_3                   7 // 3 Tasten (5 LED) Tastatur (Timeboy-Mobil-PZE usw.)

// Setzen/Ermitteln der Multimediakarte
#define IV_FUNCTION_SET_TYPE_MMC                 0x0013
#define IV_FUNCTION_GET_TYPE_MMC                 0x0014

#define IV_VALUE_TYPE_MMC_NONE                   0 // Keine MMC eingebaut
#define IV_VALUE_TYPE_MMC_EXTENSION              1 // Als Speichererweiterung
#define IV_VALUE_TYPE_MMC_DISK                   2 // Als Wechseldatentraeger
#define IV_VALUE_TYPE_MMC_USB                    3 // USB-Host als Speichererweiterung


// Setzen/Ermitteln des GPS - Typs
#define IV_FUNCTION_SET_TYPE_GPS                 0x0015
#define IV_FUNCTION_GET_TYPE_GPS                 0x0016

#define IV_VALUE_TYPE_GPS_NONE                   0 // Kein GPS
#define IV_VALUE_TYPE_GPS_GN80                   1 // Furuno GN-80
#define IV_VALUE_TYPE_GPS_5H                     2 // Lea 5H
#define IV_VALUE_TYPE_GPS_MOUSE_5H               3 // Navimouse mit Lea 5H


// Setzen/Ermitteln der USB - Komponente
#define IV_FUNCTION_SET_COMPONENT_USB            0x0017
#define IV_FUNCTION_GET_COMPONENT_USB            0x0018

// Setzen/Ermitteln der Hardwarekonfigurationsnummer
#define IV_FUNCTION_SET_HW_CONFIG                0x0019
#define IV_FUNCTION_GET_HW_CONFIG                0x001A

// Setzen/Ermitteln des Funkmodultyps
#define IV_FUNCTION_SET_TYPE_RADIO               0x001B
#define IV_FUNCTION_GET_TYPE_RADIO               0x001C

#define IV_VALUE_TYPE_RADIO_NONE                 0 // Kein Funk
#define IV_VALUE_TYPE_RADIO_WRAPTHOR             1 // Bluegiga Wrap Thor
#define IV_VALUE_TYPE_RADIO_ZIGBEE               2 // Zigbee


// Setzen/Ermitteln des Funkuhrmodultyps
#define IV_FUNCTION_SET_TYPE_DCF77               0x001D
#define IV_FUNCTION_GET_TYPE_DCF77               0x001E

#define IV_VALUE_TYPE_DCF77_NONE                 0 // Keine Funkuhr
#define IV_VALUE_TYPE_DCF77_HKW                  1 // HKW
#define IV_VALUE_TYPE_DCF77_BARCODE              2 // Barcodeleser von Timeboy intern
#define IV_VALUE_TYPE_DCF77_BARCODE_ZEBEX        3 // Barcodeleser Zebex CT-130 von Timeboy


// Setzen/Ermitteln der Option fuer analoge Eingaenge (32bit)
#define IV_FUNCTION_SET_COMPONENT_IN_ANALOG      0x001F
#define IV_FUNCTION_GET_COMPONENT_IN_ANALOG      0x0020

#define IV_VALUE_COMPONENT_IN_ANALOG_NONE        0x00000000
#define IV_VALUE_COMPONENT_IN_ANALOG_1           0x00000001
#define IV_VALUE_COMPONENT_IN_ANALOG_2           0x00000003


// Setzen/Ermitteln der Option fuer digitale Eingaenge (32bit)
#define IV_FUNCTION_SET_COMPONENT_IN_DIGITAL     0x0021
#define IV_FUNCTION_GET_COMPONENT_IN_DIGITAL     0x0022

#define IV_VALUE_COMPONENT_IN_DIGITAL_NONE       0x00000000
#define IV_VALUE_COMPONENT_IN_DIGITAL_1          0x00000001
#define IV_VALUE_COMPONENT_IN_DIGITAL_2          0x00000003
#define IV_VALUE_COMPONENT_IN_DIGITAL_3          0x00000007
#define IV_VALUE_COMPONENT_IN_DIGITAL_4          0x0000000F
#define IV_VALUE_COMPONENT_IN_DIGITAL_5          0x0000001F
#define IV_VALUE_COMPONENT_IN_DIGITAL_6          0x0000003F
#define IV_VALUE_COMPONENT_IN_DIGITAL_7          0x0000007F
#define IV_VALUE_COMPONENT_IN_DIGITAL_8          0x000000FF


// Setzen/Ermitteln der Option fuer digitale Ausgaenge (32bit)
#define IV_FUNCTION_SET_COMPONENT_OUT_DIGITAL    0x0023
#define IV_FUNCTION_GET_COMPONENT_OUT_DIGITAL    0x0024

#define IV_VALUE_COMPONENT_OUT_DIGITAL_NONE      0x00000000
#define IV_VALUE_COMPONENT_OUT_DIGITAL_1         0x00000001
#define IV_VALUE_COMPONENT_OUT_DIGITAL_2         0x00000003
#define IV_VALUE_COMPONENT_OUT_DIGITAL_3         0x00000007
#define IV_VALUE_COMPONENT_OUT_DIGITAL_4         0x0000000F


// Setzen/Ermitteln des Fingerprinttyps
#define IV_FUNCTION_SET_TYPE_FINGERPRINT         0x0025
#define IV_FUNCTION_GET_TYPE_FINGERPRINT         0x0026

#define IV_VALUE_TYPE_FINGERPRINT_NONE           0
#define IV_VALUE_TYPE_FINGERPRINT_BIOKEY_3000    1 // Idencom (BioKey 3000 - Modul)
#define IV_VALUE_TYPE_FINGERPRINT_BIOKEY_4000    2 // Idencom (BioKey 4000 - Modul)
#define IV_VALUE_TYPE_FINGERPRINT_BIOKEY_4020    3 // Idencom (BioKey 4020 - Modul)

// Setzen/Ermitteln der Typen der Externen Lesermodule
#define IV_FUNCTION_SET_TYPE_EXTERNAL_READER     0x0027
#define IV_FUNCTION_GET_TYPE_EXTERNAL_READER     0x0028

#define IV_VALUE_TYPE_EXTERNAL_READER_UNDEFINED  0x00 // Undefined
#define IV_VALUE_TYPE_EXTERNAL_READER_GIS        0x01 // GIS
#define IV_VALUE_TYPE_EXTERNAL_READER_PHG        0x02 // PHG
#define IV_VALUE_TYPE_EXTERNAL_READER_PROTECT    0x04 // Protect
#define IV_VALUE_TYPE_EXTERNAL_READER_IDTRONIC   0x08 // ID Tronic UHF

// Setzen/Ermitteln der Optionen fuer die Firmware
#define IV_FUNCTION_SET_OPTION_IOMODULE          0x0029
#define IV_FUNCTION_GET_OPTION_IOMODULE          0x002A

#define IV_VALUE_OPTION_IOMODULE_UNDEFINED       0xFF // Undefiniert
#define IV_VALUE_OPTION_IOMODULE_NONE            0x00 // Keines
#define IV_VALUE_OPTION_IOMODULE                 0x01 // IOModule Anbindung benoetigt

// Setzen/Ermitteln der Optionen fuer die Firmware
#define IV_FUNCTION_SET_OPTION_COMSERVER         0x002B
#define IV_FUNCTION_GET_OPTION_COMSERVER         0x002C

#define IV_VALUE_OPTION_COMSERVER_UNDEFINED      0xFF // Undefiniert
#define IV_VALUE_OPTION_COMSERVER_NONE           0x00 // Keines
#define IV_VALUE_OPTION_COMSERVER                0x01 // ComServerMode wird benoetigt

// Setzen/Ermitteln der Optionen fuer die Firmware
#define IV_FUNCTION_SET_OPTION_TBADAPTION        0x002D
#define IV_FUNCTION_GET_OPTION_TBADAPTION        0x002E

#define IV_VALUE_OPTION_TBADAPTION_UNDEFINED     0xFF // Undefiniert
#define IV_VALUE_OPTION_TBADAPTION_NONE          0x00 // Keines
#define IV_VALUE_OPTION_TBADAPTION               0x01 // ComServerMode wird benoetigt

// Ausfuehren von Befehlen fuer den Pruefplatz
#define IV_FUNCTION_INSPECTION_STATION           0x002F
#define IV_FUNCTION_INSPECTION_STATION_OK        0x0030

// Setzen/Ermitteln der Optionen fuer die Firmware
#define IV_FUNCTION_SET_OPTION_TPWRITE           0x0031
#define IV_FUNCTION_GET_OPTION_TPWRITE           0x0032

#define IV_VALUE_OPTION_TPWRITE_UNDEFINED        0xFF // Undefiniert
#define IV_VALUE_OPTION_TPWRITE_NONE             0x00 // Keines
#define IV_VALUE_OPTION_TPWRITE                  0x01 // wird benoetigt

// Setzen/Ermitteln der Optionen fuer die Firmware
#define IV_FUNCTION_SET_OPTION_HTTP              0x0033
#define IV_FUNCTION_GET_OPTION_HTTP              0x0034

#define IV_VALUE_OPTION_HTTP_UNDEFINED           0xFF // Undefiniert
#define IV_VALUE_OPTION_HTTP_NONE                0x00 // Keines
#define IV_VALUE_OPTION_HTTP                     0x01 // wird benoetigt

// Setzen/Ermitteln der Optionen fuer die Firmware
#define IV_FUNCTION_SET_OPTION_LANGUAGE          0x0035
#define IV_FUNCTION_GET_OPTION_LANGUAGE          0x0036

#define IV_VALUE_OPTION_LANGUAGE_UNDEFINED       0xFF // Undefiniert
#define IV_VALUE_OPTION_LANGUAGE_NONE            0x00 // Keines
#define IV_VALUE_OPTION_LANGUAGE                 0x01 // wird benoetigt

// Setzen/Ermitteln der Optionen fuer die Firmware
#define IV_FUNCTION_SET_OPTION_BIOKEY_UPDATE     0x0037
#define IV_FUNCTION_GET_OPTION_BIOKEY_UPDATE     0x0038

#define IV_VALUE_OPTION_BIOKEY_UPDATE_UNDEFINED  0xFF // Undefiniert
#define IV_VALUE_OPTION_BIOKEY_UPDATE_NONE       0x00 // Keines
#define IV_VALUE_OPTION_BIOKEY_UPDATE            0x01 // wird benoetigt

// Setzen/Ermitteln der Optionen fuer das Setup
#define IV_FUNCTION_SET_SETUP_OPTIONS            0x0039
#define IV_FUNCTION_GET_SETUP_OPTIONS            0x003A

#define IV_VALUE_SETUP_OPTION_MASK               0x0001 // Maske der unterstützten Setupoptionen
#define IV_VALUE_SETUP_OPTION_ALL                0xFFFF
#define IV_VALUE_SETUP_OPTION_IOCONTROL          0x0001 // Signalverarbeitung

// Setzen/Ermitteln des Touchtyp
#define IV_FUNCTION_SET_TYPE_TOUCH               0x003B
#define IV_FUNCTION_GET_TYPE_TOUCH               0x003C

#define IV_VALUE_TYPE_TOUCH_NONE                 0 // Kein Touch
#define IV_VALUE_TYPE_TOUCH_SMALL                1 // 4 Zoll Touchpanel
#define IV_VALUE_TYPE_TOUCH_BIG                  2 // 6 Zoll Touchpanel

// Setzen/Ermitteln des Sensortyp
#define IV_FUNCTION_SET_TYPE_SENSOR              0x003D
#define IV_FUNCTION_GET_TYPE_SENSOR              0x003E

#define IV_VALUE_TYPE_SENSOR_NONE                 0 // Kein Sensor
#define IV_VALUE_TYPE_SENSOR_VIBRATION            1 // Vibrationssensor bestückt
#define IV_VALUE_TYPE_SENSOR_SPEEDUP              2 // Beschleunigungssensor bestückt

// Setzen/Ermitteln der Komponente Lithium Batterie bestückt
#define IV_FUNCTION_SET_COMPONENT_BATTERY        0x003F
#define IV_FUNCTION_GET_COMPONENT_BATTERY        0x0040

// Setzen/Ermitteln der Komponente Ladeschaltung bestückt
#define IV_FUNCTION_SET_COMPONENT_CHARGE         0x0041
#define IV_FUNCTION_GET_COMPONENT_CHARGE         0x0042

// Setzen/Ermitteln eines Extension Bords
//#define IV_FUNCTION_SET_EXTENSION_BOARDS         0x0043
//#define IV_FUNCTION_GET_EXTENSION_BORADS         0x0044

// Ermitteln der gesamten Hardwareinfo in einer einzelnen Struktur
#define IV_FUNCTION_GET_HARDWARE_INFO            0x0045

//#############################################################################
// Da es kein Protokollmodul gibt werden die wenigen benoetigten Funktionen 
// aus dem Hardwaremodul heraus genommen. Das Define ist in dem Protocolmodul zu suchen.
//#############################################################################
//#define IV_FUNCTION_EXCHANGE_PUBLIC_KEY          0x0046

// Setzen/Ermitteln der Bezahlbaren Optionen
#define IV_FUNCTION_SET_PAYABLE_OPTIONS          0x0047
#define IV_FUNCTION_GET_PAYABLE_OPTIONS          0x0048

#define IV_VALUE_PAYABLE_OPTION_MASK             0x01 // Maske der unterstützten Optionen
#define IV_VALUE_PAYABLE_OPTION_ALL              0xFF
#define IV_VALUE_PAYABLE_OPTION_DATA_ON_CARD     0x01 // Data on Card

class Channel4ModuleHardware : public Channel4ProtocolCrypt
{
public:
    Channel4ModuleHardware( short channelId = IV_DEFAULT_CHANNELID, short deviceId = IV_DEFAULT_DEVICEID );
    virtual ~Channel4ModuleHardware();

    StructHardwareInfo _hwInfo;

public:
    // Gibt Modulfehler zurueck.
    virtual void getErrorModule( short module, short error, std::string &text );
    // Gibt Funktionsfehlertext zurueck.
    virtual void getErrorFunction( short module, short func, short error, std::string &text );

    // Einfache Anfrage ob Geraet mit Protokollversion 4 erreichbar.
    int request( short deviceId );

    // MODULE_HARDWARE

    // Einfache Anfrage ob das Geraet antwortet.
    int pingRandom( short deviceId );

    // BUG:
    // Im Bootloader 00.04.00.00 && 00.04.00.01 liefert die Funktion
    // - TypeTransponder den Wert fuer das Display
    // - TypeDisplay den Wert fuer das Keypad
    // - TypeKeypad den Wert fuer den Transponder

    // Typ schreiben/lesen
    int setTypeDevice( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeDevice( short deviceId, unsigned char &val );
    int setTypeBoard( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeBoard( short deviceId, unsigned char &val );
    int setTypeTCPIP( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeTCPIP( short deviceId, unsigned char &val );
    int setTypeModemMC( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeModemMC( short deviceId, unsigned char &val );
    int setTypeDisplay( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF, int checked = FALSE );
    int getTypeDisplay( short deviceId, unsigned char &val, int checked = FALSE );
    int setTypeKeypad( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF, int checked = FALSE );
    int getTypeKeypad( short deviceId, unsigned char &val, int checked = FALSE );
    int setTypeTransponder( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF, int checked = FALSE );
    int getTypeTransponder( short deviceId, unsigned char &val, int checked = FALSE );
    int setTypeMMC( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeMMC( short deviceId, unsigned char &val );
    int setTypeRadio( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeRadio( short deviceId, unsigned char &val );
    int setTypeGPS( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeGPS( short deviceId, unsigned char &val );
    int setTypeDCF77( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeDCF77( short deviceId, unsigned char &val );
    int setTypeFingerprint( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeFingerprint( short deviceId, unsigned char &val );
    int setTypeExternalReader( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeExternalReader( short deviceId, unsigned char &val );
    int setTypeTouch( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeTouch( short deviceId, unsigned char &val );
    int setTypeSensor( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getTypeSensor( short deviceId, unsigned char &val );

    // Komponente schreiben/lesen
    int setComponentRS485_0( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getComponentRS485_0( short deviceId, unsigned char &val );
    int setComponentRS485_1( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getComponentRS485_1( short deviceId, unsigned char &val );
    int setComponentUSB( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getComponentUSB( short deviceId, unsigned char &val );
    int setComponentInAnalog( short deviceId, unsigned int val, unsigned short val2 = 0xFFFF );
    int getComponentInAnalog( short deviceId, unsigned int &val );
    int setComponentInDigital( short deviceId, unsigned int val, unsigned short val2 = 0xFFFF );
    int getComponentInDigital( short deviceId, unsigned int &val );
    int setComponentOutDigital( short deviceId, unsigned int val, unsigned short val2 = 0xFFFF );
    int getComponentOutDigital( short deviceId, unsigned int &val );
    int setComponentBattery( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getComponentBattery( short deviceId, unsigned char &val );
    int setComponentCharge( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getComponentCharge( short deviceId, unsigned char &val );

    // Optionen schreiben/lesen
    int setOptionIOModule( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getOptionIOModule( short deviceId, unsigned char &val );
    int setOptionComServer( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getOptionComServer( short deviceId, unsigned char &val );
    int setOptionTBAdaption( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getOptionTBAdaption( short deviceId, unsigned char &val );
    int setOptionTPWrite( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getOptionTPWrite( short deviceId, unsigned char &val );
    int setOptionHTTP( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getOptionHTTP( short deviceId, unsigned char &val );
    int setOptionLanguage( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getOptionLanguage( short deviceId, unsigned char &val );
    int setOptionBiokeyUpdate( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getOptionBiokeyUpdate( short deviceId, unsigned char &val );
    int setSetupOptions( short deviceId, unsigned short mask, unsigned short options );
    int getSetupOptions( short deviceId, unsigned short &mask, unsigned short &options );
    int setPayableOptions( short deviceId, unsigned char mask, unsigned char options, unsigned short val2 = 0xFFFF );
    int getPayableOptions( short deviceId, unsigned char &mask, unsigned char &options );

    // Liest die gesamte Hardwareinfo struktur aus.
    int getHardwareInfo( short deviceId, StructHardwareInfo &hwInfo );

    // HWConfigversion schreiben/lesen
    int setHWConfig( short deviceId, unsigned char val, unsigned short val2 = 0xFFFF );
    int getHWConfig( short deviceId, unsigned char &val );

    // Befehl an Pruefplatz senden
    int commandInspectionStation( short deviceId, unsigned char address, char *data, int sizeData, unsigned char &addressAnswer, char *dataAnswer, int &sizeAnswer, int toPending );
    int commandInspectionStationOK( short deviceId );

    //#################################
    // CallRoutinen

    // Ausfuehren von Befehlen
    virtual int remoteCall( short module, short func, unsigned char *data, int &count );
};

#endif
