/**
* @file  FileHardwareInfoItems.h
* @brief Schnittstelle fuer die Hilfsklassen zu FileHardwareInfo.
*/
//////////////////////////////////////////////////////////////////////
#ifndef FILEHARDWAREINFOITEMS_H
#define FILEHARDWAREINFOITEMS_H

#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif
// Strukturdefinitionen
#include "../Common/hardwareinfostructs.h"

// Verwaltung der Strukturen in Baum.
#include <string>
#include <vector>
#include <array>
#include <set>
// Leider nicht mit VS2012
//#include <initializer_list>
#include <boost/assign/list_of.hpp>

enum EnumItemHardwareInfoType
{
    Root = -1,
    MainBoard = 0,
    Module,
    ModuleWithPlaces,
    SubModule
};

// ########################################################
// Hauptkommunikation

enum EnumModuleRS232Communication
{
    // MasterIV - Gerte besitzen immer einen SerialPort
    SerialPort = 16 // M016 Serial Port SubD
};
typedef std::set<EnumModuleRS232Communication> setModuleRS232Communication;
static const setModuleRS232Communication moduleRS232Communication_all = boost::assign::list_of
                                                 (SerialPort);

enum EnumModuleRS485Communication
{
    RS485Old = 100000, // RS485
    RS485New = 35      // M035 RS485 Com. Port
};
typedef std::set<EnumModuleRS485Communication> setModuleRS485Communication;
static const setModuleRS485Communication moduleRS485Communication_all = boost::assign::list_of
                                                 (RS485Old)
                                                 (RS485New);

enum EnumModuleUSBCommunication
{
    USBOld = 100004 // USB
    // EVO-Line - Gerte besitzen immer USB.
};
typedef std::set<EnumModuleUSBCommunication> setModuleUSBCommunication;
static const setModuleUSBCommunication moduleUSBCommunication_all = boost::assign::list_of
                                                 (USBOld);

enum EnumModuleLANCommunication
{
    LANOld = 100001, // TCPIP-XPort
    LANNew = 11      // M011 Ethernet Port
};
typedef std::set<EnumModuleLANCommunication> setModuleLANCommunication;
static const setModuleLANCommunication moduleLANCommunication_all = boost::assign::list_of
                                                 (LANOld)
                                                 (LANNew);

enum EnumModuleWLANCommunication
{
    WLANOld = 100002, // WLAN-MatchPort
    WLANNew1 = 1,     // M001 WLAN Redpine RS9110
    WLANNew2 = 105002  // WLAN RS9110-N-11-24
};
typedef std::set<EnumModuleWLANCommunication> setModuleWLANCommunication;
static const setModuleWLANCommunication moduleWLANCommunication_all = boost::assign::list_of
                                                 (WLANOld)
                                                 (WLANNew1)
                                                 (WLANNew2);

enum EnumModuleMobileCommunication
{
    MC35_39 = 100003, // Mobile MC35-39
    MC55    = 105001  // Mobile MC55i
                      // Sonderfall: Wenn das Aufltmodul M015 bestckt wurde und kein Ansteckmodul angegeben ist, wird von einem MC55 augegangen.
};
typedef std::set<EnumModuleMobileCommunication> setModuleMobileCommunication;
static const setModuleMobileCommunication moduleMobileCommunication_all = boost::assign::list_of
                                                 (MC35_39)
                                                 (MC55);

enum EnumModuleBluetoothCommunication
{
    Bluegiga = 100005 // Bluegiga Wrap Thor
};
typedef std::set<EnumModuleBluetoothCommunication> setModuleBluetoothCommunication;
static const setModuleBluetoothCommunication moduleBluetoothCommunication_all = boost::assign::list_of
                                                 (Bluegiga);

enum EnumModuleMMCCommunication
{
    MMCRemovable = 100009 // Removable media device
};
typedef std::set<EnumModuleMMCCommunication> setModuleMMCCommunication;
static const setModuleMMCCommunication moduleMMCCommunication_all = boost::assign::list_of
                                                 (MMCRemovable);

enum EnumModuleUSBHostCommunication
{
    JackPlug = 24,     // M024 USB 2.0 A Jack-Plug
    Vinculum = 100010  // USB-Host Vinculum
};
typedef std::set<EnumModuleUSBHostCommunication> setModuleUSBHostCommunication;
static const setModuleUSBHostCommunication moduleUSBHostCommunication_all = boost::assign::list_of
                                                 (JackPlug)
                                                 (Vinculum);

// ########################################################
// Zusatzbaugruppen

enum EnumModuleAccessControl
{
    AccessControlOld = 100006, // RS485 fr ZK.
    AccessControlNew = 14      // M014 RS485 + 12V Supply
};
typedef std::set<EnumModuleAccessControl> setModuleAccessControl;
static const setModuleAccessControl moduleAccessControl_all = boost::assign::list_of
                                                 (AccessControlOld)
                                                 (AccessControlNew);

enum EnumModuleBiokey
{
    Biokey3000 = 104000, // Biokey 3000
    Biokey4000 = 104001, // Biokey 4000
    Biokey4020 = 104002  // Biokey 4020
};
typedef std::set<EnumModuleBiokey> setModuleBiokey;
static const setModuleBiokey moduleBiokey_all = boost::assign::list_of
                                                 (Biokey3000)
                                                 (Biokey4000)
                                                 (Biokey4020);


enum EnumModuleGPS
{
    Mod01_007 = 7,      // GPS_Mod01_007
    Furuno    = 100011, // Furuno GN-80
    Lea5H     = 100012, // Lea5H
    Navimouse = 100013  // Navimouse mit Lea5H
};
typedef std::set<EnumModuleGPS> setModuleGPS;
static const setModuleGPS moduleGPS_all = boost::assign::list_of
                                                 (Mod01_007)
                                                 (Furuno)
                                                 (Lea5H)
                                                 (Navimouse);

enum EnumModuleDisplay
{
    Monocrome_3_5 = 110001, // Monocrome Display 3,5"
                            // Sonderfall: Wenn das Aufltmodul M018 bestckt wurde und kein Ansteckmodul angegeben ist, wird von einem Monocrome_3_5 augegangen.
    Color_1_9     = 110002, // Color TFT 1,9" 220x176
    Color_2_8     = 110003, // Color TFT 2,8" 320x240
                            // Sonderfall: Wenn das Aufltmodul M030 bestckt wurde und kein Ansteckmodul angegeben ist, wird von einem Color_2_8 augegangen.
    Color_4_3     = 110004, // Color TFT 4,3" 420x272
                            // Sonderfall: Wenn das Aufltmodul M019 bestckt wurde und kein Ansteckmodul angegeben ist, wird von einem Color_4_3 augegangen.
    Color_2_0     = 110005  // Color TFT 2,0" 220x176
};
typedef std::set<EnumModuleDisplay> setModuleDisplay;
static const setModuleDisplay moduleDisplay_all = boost::assign::list_of
                                                    (Monocrome_3_5)
                                                    (Color_1_9)
                                                    (Color_2_8)
                                                    (Color_4_3)
                                                    (Color_2_0);

enum EnumModuleTouch
{
    Touch_Evo_4_3 = 110101, // Touchscreen + SMT-C-T10I
                            // Sonderfall: Wenn das Aufltmodul M019 bestckt wurde und kein Ansteckmodul angegeben ist, wird von einem Touch_Evo_4_3 augegangen.
    Touch_Evo_2_8 = 110102  // Resistive Touch 3.5"
                            // Sonderfall: Wenn das Aufltmodul M030 bestckt wurde und kein Ansteckmodul angegeben ist, wird von einem Touch_Evo_2_8 augegangen.
};
typedef std::set<EnumModuleTouch> setModuleTouch;
static const setModuleTouch moduleTouch_all = boost::assign::list_of
                                                (Touch_Evo_4_3)
                                                (Touch_Evo_2_8);


class ItemBoard;

typedef std::set<ItemBoard *> setItemBoard;
typedef std::set<unsigned int> setModuleId;

class ItemHardwareInfoBase
{
public:
    ItemHardwareInfoBase( ItemHardwareInfoBase *parent, EnumItemHardwareInfoType typeId );
    virtual ~ItemHardwareInfoBase();

    ItemHardwareInfoBase *parent();
    void setParent( ItemHardwareInfoBase *parent );

    EnumItemHardwareInfoType typeId();
    void setTypeId( EnumItemHardwareInfoType typeId );

private:
    EnumItemHardwareInfoType _typeId;
    ItemHardwareInfoBase *_parent;

public:
    void setAutoCorrectModuleDisabled( bool state );
    bool autoCorrectModuleDisabled();

private:
    static bool _autoCorrectModuleDisabled;
};

class ItemJumper
{
public:
    ItemJumper();
    virtual ~ItemJumper();

public:
    void reset();
    bool writeSerialize( std::vector<unsigned char>::iterator iter );
    bool readSerialize( std::vector<unsigned char>::const_iterator iter );

    HARDWARECONFIGJUMPER &hcj();

protected:
    std::array<unsigned char, sizeof(HARDWARECONFIGJUMPER)> _structData;

public:
    std::string toJson();
};

class ItemBoard : public ItemHardwareInfoBase
{
public:
    ItemBoard( ItemHardwareInfoBase *parent = 0, EnumItemHardwareInfoType typeId = MainBoard );
    virtual ~ItemBoard();

public:
    // Tabelle der Id's, Stand: 03.02.2014
    // Modul Typ 0 Hauptplatinen Nummernkreis 50.000 bis 99.999
    // Modul Typ 1 und 2  Aufltmodule im Datafox Standard Raster Nummernkreis 1 bis 49.999
    // Modul Typ 3 Ansteckmodule fr Modulplatinen Typ 0 und 2 Nummernkreis ab 100.000
    enum BoardIdGroup
    {
        Standard                    = 1,
        StandardEnd                 = 49999,
        Hauptplatine                = 50000,
        HauptplatineEnd             = 99999,
        Tastatur                    = 101000,
        TastaturEnd                 = 101999,
        Transponderleser            = 102000,
        TransponderleserEnd         = 102999,
        Barcode                     = 103000,
        BarcodeEnd                  = 103999,
        Fingerprint                 = 104000,
        FingerprintEnd              = 104999,
        Kommunikation               = 105000,
        KommunikationEnd            = 105999,
        Leuchtflaechen              = 106000,
        LeuchtflaechenEnd           = 106999,
        Displays                    = 110000,
        DisplaysEnd                 = 110099,
        Touch                       = 110100,
        TouchEnd                    = 110199,
    };

    // Tabelle der Id's, Stand: 01.12.2014
    enum BoardId
    {
        // Hauptplatinen
        MainboardIPCExtensionA                             = 50003, // IPC Vario 5.7 Mainboard
        MainboardIPCStandard                               = 50005, // IPC Standard Mainboard
        MainboardCombiV4                                   = 50007, // Kombi V4 Mainboard
        MainboardIOBoxTerminal                             = 50006, // IO-Box Terminal
        MainboardDockingstation                            = 50008, // Dockingstation Mainboard
        MainboardIOBox                                     = 50009, // IO-Box Top Hat Rail
        MainboardCombiV1_3                                 = 50014, // Kombi V1-3 Mainboard
        MainboardBDEMasterIV                               = 50015, // BDE-MasterIV Mainboard
        MainboardTimeboy4                                  = 50016, // Timeboy-V4 Mainboard

        // Aufltmodule
        ModuleWLAN                                         =  1, // M001 WLAN Redpine RS9110
        ModuleChargeRegulator                              =  4, // M004 Charge Regulator
        ModuleDO                                           =  5, // M005 Relay Output
        ModuleDI                                           =  6, // M006 Digital Input
        ModuleGPS                                          =  7, // M007 GPS LEA6-H
        ModuleAI                                           =  8, // M008 Analog Input
        Module009                                          =  9, // M009 Mobil Master Port
        Module010                                          = 10, // M010 Backlight Control
        Module011                                          = 11, // M011 Ethernet Port
        ModuleDIDO                                         = 12, // M012 Digital In-/Output
        Module013                                          = 13, // M013 Position Receiver
        Module014                                          = 14, // M014 RS485 + 12V Supply
        Module015                                          = 15, // M015 Mobile Phone Plug
        Module016                                          = 16, // M016 Serial Port SubD
        Module017                                          = 17, // M017 Micro SD Card
        Module018                                          = 18, // M018 Grafic Adapter
        Module019                                          = 19, // M019 Grafic Adapter
        Module020                                          = 20, // M020 Touch Key
        Module021                                          = 21, // M021 Double Serial Port
        Module024                                          = 24, // M024 USB 2.0 A Jack-Plug
        Module030                                          = 30, // M030 Grafic Adapter
        Module032                                          = 32, // M032 Serial Port M.-DIN
        Module035                                          = 35, // M035 RS485 Com. Port
        Module037                                          = 37, // M037 Single Serial Port

        ModuleOldAI                                        = 100016,
        ModuleOldDI                                        = 100017,
        ModuleOldDO                                        = 100018

    };
public:
    void reset();
    bool writeSerialize( std::vector<unsigned char> &data );
    bool readSerialize( std::vector<unsigned char>::const_iterator iterRead, const std::vector<unsigned char> &data );

    HARDWARECONFIG &hc();

    typedef std::vector<ItemBoard> typeChilds;
    typedef std::vector<ItemJumper> typeJumpers;

public:
    setItemBoard collectModuleIds( const setModuleId &ids );
    static std::string idGroup( unsigned int id );

protected:
    std::array<unsigned char, sizeof(HARDWARECONFIG)> _structData;
    typeChilds _childs;
    typeJumpers _jumpers;
    void getDefault( ItemBoard &item );

public:
    std::string toJson();
    typeChilds &childs();
    typeJumpers &jumpers();
};

class ItemDeviceHardwareInfo : public ItemHardwareInfoBase
{
public:
    ItemDeviceHardwareInfo( ItemHardwareInfoBase *parent = 0, EnumItemHardwareInfoType typeId = Root );
    virtual ~ItemDeviceHardwareInfo();

public:
    void reset();
    bool writeSerialize( std::vector<unsigned char> &buffer );
    bool readSerialize( const std::vector<unsigned char> &buffer );

    HARDWARECONFIGNODE &hcn();

public:
    setItemBoard collectModuleIds( const setModuleId &ids );

    setItemBoard moduleRS232Communication( const setModuleRS232Communication &module = moduleRS232Communication_all );
    setItemBoard moduleRS485Communication( const setModuleRS485Communication &module = moduleRS485Communication_all );
    setItemBoard moduleUSBCommunication( const setModuleUSBCommunication &module = moduleUSBCommunication_all );
    setItemBoard moduleLANCommunication( const setModuleLANCommunication &module = moduleLANCommunication_all );
    setItemBoard moduleWLANCommunication( const setModuleWLANCommunication &module = moduleWLANCommunication_all );
    setItemBoard moduleMobileCommunication( const setModuleMobileCommunication &module = moduleMobileCommunication_all );
    setItemBoard moduleBluetoothCommunication( const setModuleBluetoothCommunication &module = moduleBluetoothCommunication_all );
    setItemBoard moduleMMCCommunication( const setModuleMMCCommunication &module = moduleMMCCommunication_all );
    setItemBoard moduleUSBHostCommunication( const setModuleUSBHostCommunication &module = moduleUSBHostCommunication_all );

    setItemBoard moduleAccessControl( const setModuleAccessControl &module = moduleAccessControl_all );
    setItemBoard moduleBiokey( const setModuleBiokey &module = moduleBiokey_all );
    setItemBoard moduleGPS( const setModuleGPS &module = moduleGPS_all );

    setItemBoard moduleDisplay( const setModuleDisplay &module = moduleDisplay_all );
    setItemBoard moduleTouch( const setModuleTouch &module = moduleTouch_all );

protected:
    std::array<unsigned char, sizeof(HARDWARECONFIGNODE)> _structData;
    ItemBoard::typeChilds _childs;

public:
    std::string toJson();
    ItemBoard::typeChilds &childs();
};

#endif
