/**
 * @file  DFLog.cpp
 * @brief Implementierung der Klasse CDFLog.
 */
//////////////////////////////////////////////////////////////////////

#include "DFLog.h"

#ifdef _MSC_VER
#include <windows.h>
#include <direct.h>
#include <io.h>
#else
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <stdio.h>
#include <stdarg.h>
#endif

CDFLog::CDFLog(short channelID)
{
#ifdef _MSC_VER
    // Initialize the critical section one time only.
    InitializeCriticalSection( &m_mutexFile ); 
#else
    pthread_mutex_init( &m_mutexFile, NULL );
#endif

    // Zugewiesene Kanalnummer
    m_channelID = channelID;
    m_logCheck.m_channelID = channelID;

    // Bevorzugt auf die Dockingstation Type 1 prüfen.
    _dockingPreferType = 1;

    // Initialisierungsdatei lesen
    logIniRead();
    logCheck();
}

CDFLog::~CDFLog()
{
#ifdef _MSC_VER
    // Release resources used by the critical section object.
    DeleteCriticalSection( &m_mutexFile );
#else
    pthread_mutex_destroy( &m_mutexFile );
#endif
}

void CDFLog::logIniRead()
{
    char szTmp[260];
    char szFileName[260];
    memset(szFileName, 0, sizeof(szFileName));

#ifdef _MSC_VER
    // Dateiname der Inidatei
    GetModuleFileNameA(NULL, szFileName, sizeof(szFileName));
    
    MakeFilePath(szFileName, "DFCom", "ini");
#else
    char szBasePath[260];
    memset(szBasePath, 0 , sizeof(szBasePath));

    sprintf(szTmp, "/proc/%d/exe", getpid());  
    readlink(szTmp, szBasePath, sizeof(szBasePath));  
    
    char *p = &szBasePath[strlen(szBasePath)];
    while (p != szBasePath && *(p - 1) != '/') p--;
    *p = 0;
    
    sprintf(szFileName, "%sDFCom.ini", szBasePath);
#endif
    
    // pruefen und evnt. anlegen der Inidatei
    struct stat lstat;
    if (stat(szFileName, &lstat) != 0) {
        FILE *fp;
        if ((fp = fopen(szFileName, "w")) == (FILE*)0) {
            // Fehler bei Erzeugung
            m_bEnable = false;
            m_bEraseFiles = false;
            return;
        }

        // Sektion "Log" anlegen
        fprintf(fp, "[Log]\n");
#ifndef _MSC_VER
        fprintf(fp, "Enable=0\n"); // Unter Linux ist das Loggen standardmaessig aus.
#endif
#ifdef _MSC_VER
        fprintf(fp, "Enable=1\n");
        // Aktuelles Verzeichnis verwenden
        memset(szTmp, 0, sizeof(szTmp));
        strcpy( szTmp, szFileName );
        MakeFilePath(szTmp, "", "");
        fprintf(fp, "Path=");
        for (int i=0; szTmp[i]; i++) {
            fprintf(fp, "%c", szTmp[i]);
            if (szTmp[i] == '\\') { // Maskierungszeichen einfuegen
                fprintf(fp, "\\");
            }
        }
        fprintf(fp, "\n");
#else
        if (szBasePath[0] == '\0')
        {
          fprintf(fp, "Path=./\n");
        }
        else
        {
          fprintf(fp, "Path=%s\n", szBasePath);
        }
#endif
        fprintf(fp, "Folder=%s\n", SUB_FOLDER);
        fprintf(fp, "Trace=0\n");
        fprintf(fp, "Erase=1\n");
        fprintf(fp, "Keep=10\n");
        fprintf(fp, "Elapse=1\n");
        fprintf(fp, "List=1\n");

        fprintf(fp, "\n\n");
        fprintf(fp, "; Kurzbeschreibung der enthaltenen Parameter:\n");
        fprintf(fp, "; Enable: Aktivieren des Hintergrundloggens. Wertebereich 0 = Nicht loggen, 1 = Loggen aktiviert.\n");
        fprintf(fp, "; Path: Gibt das Basisverzeichnis an. Bei Verwendung von Backslash-Zeichen \\ muessen diese doppelt eingegeben werden.\n");
        fprintf(fp, "; Folder: Das zu verwendende Unterverzeichnis, dieses wird wenn nicht bereits vorhanden angelegt.\n");
        fprintf(fp, "; Trace: Gibt an ob das DLL eigene Detaillog aktiviert sein soll. Wertebereich ist 1 = Ja aktivieren, 2 = Ja aktivieren mit Packetdaten, 0 = Deaktiviert.\n; \tBitte beachten Sie, dass hierbei groessere Datenmengen anfallen.\n");
        fprintf(fp, "; Erase: Gibt an ob die Logdateien automatisch geloescht werden unter Linux werden die Logdateien nicht automatisch geloescht. Wertebreich ist 1 = Ja automatisch loeschen, 0 = Nein manuell l?schen.\n; \tBitte beachten  Sie dass die Dateiverwaltung nach Abschalten uebernommen werden muss.\n");
        fprintf(fp, "; Keep: Gibt bei Erase=1 die Anzahl der zu haltenden Logtage an (nicht unter Linux gueltig). Wertebereich ist 1 - 100 Logtage. Logtage bedeutet hier, Tage an denen geloggt wurde.\n");
        fprintf(fp, "; Elapse: Gibt an ob die Ausf?hrungsdauer der jeweiligen Funktion in eckigen Klammern hinter jedem Logeintrag ausgegeben wird, dabei hat dieser folgende Aufteilung [Sekunden.Hundertstel].\n");
        fprintf(fp, "; List: Gibt an ob die uebertragenen Listendaten vor jedem uebertragen, in das Logverzeichnis exportiert werden.\n; \tWertebereich ist 0 = Nein, 1 = In Binaerformat mitprotokollieren, 2 = In Binaerform und Textformat (tab-Getrennt).\n");

        // Sektion "Docking" anlegen
        fprintf(fp, "\n\n");
        fprintf(fp, "[Docking]\n");
        fprintf(fp, "Prefer=1\n");
        fprintf(fp, "Autodetect=1\n");

        fprintf(fp, "\n\n");
        fprintf(fp, "; Kurzbeschreibung der enthaltenen Parameter:\n");
        fprintf(fp, "; Sektion [Docking] ist global gültig. Für eine Kanalspezifische Einstellung verwenden Sie bitte\n");
        fprintf(fp, "; einen Sektionsnamen [Docking_channelId] z. B. [Docking_5] für die Kanalnummer 5.\n");
        fprintf(fp, "; Prefer: Bevorzugte Version der Dockingstation. Wertebereich 1 = Version 1, 2 = Version 2.\n");
        fprintf(fp, "; Autodetect: Automatisch die Version der Dockingstation ermitteln, angefangen wird die Ermittlung mit dem Wert von Prefer. Wertebereich 0 = Aus, 1 = An.\n");

        fclose(fp);
    }

    // Log aktiv?
    m_bEnable = (m_logIni.GetUInt("Log", "Enable", 1, szFileName)) ? true : false;
    if (!m_bEnable) {

        // Abbruch durch Deaktivierung.
        m_traceMode = 0; // Dann auch Trace deaktivieren.
        m_bEraseFiles = false;
        m_bWriteList = 0;
        return;
    }
    // Trace aktiv?
    m_traceMode = m_logIni.GetUInt("Log", "Trace", 0, szFileName);

    // Basispfad fuer Log
    memset(m_logCheck.m_szBasePath, 0, sizeof(m_logCheck.m_szBasePath));
    m_logIni.GetString("Log", "Path", m_logCheck.m_szBasePath, szFileName);
    if (m_logCheck.m_szBasePath[0] == 0) 
    {
#ifdef _MSC_VER
        // Aktuelles Anwendungsverzeichnis
        strcpy( m_logCheck.m_szBasePath, szFileName );
        MakeFilePath(m_logCheck.m_szBasePath, "", "");
#else
        //getcwd(m_logCheck.m_szBasePath, sizeof(m_logCheck.m_szBasePath));
        strcpy(m_logCheck.m_szBasePath, szBasePath);
#endif
    } 
    else 
    {
#ifdef _MSC_VER
        if (m_logCheck.m_szBasePath[strlen(m_logCheck.m_szBasePath) - 1] != '\\') 
        {
            strcat(m_logCheck.m_szBasePath, "\\");
        }
#else
        if (m_logCheck.m_szBasePath[strlen(m_logCheck.m_szBasePath) - 1] != '/') 
        {
            strcat(m_logCheck.m_szBasePath, "/");
        }
#endif
    }
    // Unterordner fuer Logdateien
    memset(m_logCheck.m_szSubFolder, 0, sizeof(m_logCheck.m_szSubFolder));
    m_logIni.GetString("Log", "Folder", m_logCheck.m_szSubFolder, szFileName);
    if (m_logCheck.m_szSubFolder[0] == 0) {
        strcpy(m_logCheck.m_szSubFolder, SUB_FOLDER);
    }
    sprintf(szTmp, "%s%s", m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder);
    // pruefen und evnt. anlegen des Ordners
    if (stat(szTmp, &lstat) != 0) {
#ifdef _MSC_VER
        if (mkdir(szTmp) != 0) {
#else
        if (mkdir(szTmp, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) != 0) {
#endif
            // Fehler bei Erzeugung
            m_bEnable = false;
            m_bEraseFiles = false;
            m_bWriteList = 0;
            return;
        }
    }
#ifdef _MSC_VER
    // Schreibrechte prüfen.
    if ( _access( szTmp, 2 ) == -1 )
    {
        // Fehler bei Erzeugung
        m_bEnable = false;
        m_bEraseFiles = false;
        m_bWriteList = 0;
        return;
    }
    if (m_logCheck.m_szSubFolder[strlen(m_logCheck.m_szSubFolder) - 1] != '\\') {
        strcat(m_logCheck.m_szSubFolder, "\\");
    }
#else
    if (m_logCheck.m_szSubFolder[strlen(m_logCheck.m_szSubFolder) - 1] != '/') {
        strcat(m_logCheck.m_szSubFolder, "/");
    }
#endif
    // Dateiname der Logdatei
    time_t t = time(NULL);
    tm *ltm = localtime(&t);
    sprintf(m_szFileName, LOG_FILENAME, m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder, 
        ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_channelID);
    sprintf(m_szFileNameTrace, LOG_FILENAME, m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder, 
        ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_channelID + 1000);
    // Setzen des Namens fuer Tracefile in abgeleiteter Klasse
    setTraceFileName(m_szFileNameTrace);
    // Alte Dateien loeschen
    m_bEraseFiles = (m_logIni.GetUInt("Log", "Erase", 1, szFileName)) ? true : false;
    m_logCheck.m_eraseFileDays = m_logIni.GetUInt("Log", "Keep", 10, szFileName);
    if (m_logCheck.m_eraseFileDays < 1) m_logCheck.m_eraseFileDays = 1;
    if (m_logCheck.m_eraseFileDays > 100) m_logCheck.m_eraseFileDays = 100;
    // Ausfuehrungszeit angeben
    m_bElapse = (m_logIni.GetUInt("Log", "Elapse", 1, szFileName)) ? true : false;
    // Listen wegschreiben
    m_bWriteList = m_logIni.GetUInt("Log", "List", 1, szFileName);

    char section[20];
    sprintf( section, "Docking_%i", m_channelID );
    // Bevorzugt Dockingstation V1/2 ansprechen.
    _dockingPreferType = m_logIni.GetUInt( section, "Prefer", 1234, szFileName );
    if ( _dockingPreferType == 1234 ) // Ggf. globale Einstellung laden.
        _dockingPreferType = m_logIni.GetUInt( "Docking", "Prefer", 1, szFileName );

    if ( _dockingPreferType < 1 )
        _dockingPreferType = 1;
    else if ( _dockingPreferType > 2 )
        _dockingPreferType = 2;

    // Automatisch die Dockingstation V1/2 ermitteln.
    _dockingAutoDetect = m_logIni.GetUInt( section, "Autodetect", 1234, szFileName );
    if ( _dockingAutoDetect == 1234 )
        _dockingAutoDetect = m_logIni.GetUInt( "Docking", "Autodetect", 1, szFileName );

    if ( _dockingAutoDetect < 1 )
        _dockingAutoDetect = 0;
    else if ( _dockingAutoDetect > 0 )
        _dockingAutoDetect = 1;
}

void CDFLog::logCheck() 
{
    // Soll durchgefuehrt werden?
    if (!m_bEnable || !m_bEraseFiles) {
        
        return;
    }

#ifdef _MSC_VER
/*
    // Suche nach Dateien
    sprintf(szFile, LOG_SEARCH_FILENAME, m_szBasePath, m_szSubFolder);
    intptr_t fp;
    struct finddata_t finddata;
    if ((fp = findfirst(szFile, &finddata)) != -1) {

        do {

            for (i=0; i < 5; i++) {
                // Stimmt mit einem der zu erhaltenden Dateien ueberein?
                if (strstr(szFileName[i], finddata.name) != 0) {
                    break;
                }
            }
            if (i == 5) {
                // Sammeln der zu loeschenden Dateien,
                // da erst nach findclose geloescht werden kann!
                strcpy(szDeleteFileName[j++], finddata.name);
            } 

        } while ((j < 5) && (findnext(fp, &finddata) != -1));
        findclose(fp);

        if (j) {
            // Dateien loeschen
            for (i=0; i < j; i++) {
            
                sprintf(szFile, LOG_DELETE_FILENAME, m_szBasePath, m_szSubFolder, szDeleteFileName[i]);
                if (remove(szFile) == -1) {
                    // Konnte nicht geloescht werden!
                }
            }
        }
    }
*/
    m_logCheck.Check();

#else // 2011_02_03: VORSICHT: Unter Linux werden die Log-Dateien nicht gelöscht.
/*
    int i, j;
    char szFile[260];
    char szFileName[5][260];
    char szDeleteFileName[5][260];

    // Namen der Logdateien fuer die letzten fuenf Tage erstellen.
    for (i=0; i < 5; i++) {
        // Dateiname der Logdatei
        time_t t = time(NULL);
        t -= 3600 * 24 * i;
        tm *ltm = localtime(&t);
        sprintf(szFileName[i], LOG_FILENAME, m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder, 
            ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_channelID);
        
        //@LOG:
        printf("Halte Datei: %s\n", szFileName[i]);
    }
    // Alle anderen Logdateien loeschen 

    j = 0;

    // Suche nach Dateien
    sprintf(szFile, "%s%s", m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder);

    DIR *dir;
    if ((dir = opendir(szFile)) != NULL) {
        
        struct dirent *ent;
        while((j < 5) && ((ent = readdir(dir)) != NULL)) {

            if (ent->d_type != DT_REG && ent->d_type != DT_UNKNOWN)
            {
              continue;
            }
            for (i=0; i < 5; i++) {
                // Stimmt mit einem der zu erhaltenden Dateien ueberein?
                if (strstr(szFileName[i], ent->d_name) != 0) {
                    //@LOG:
                    printf("Datei wird gehalten: %s\n", szFileName[i]);
                    break;
                }
            }
            if (i == 5) {
                // Sammeln der zu loeschenden Dateien,
                // da erst nach closedir geloescht werden kann!
                strcpy(szDeleteFileName[j++], ent->d_name);
            } 
        }
        closedir(dir);

        if (j) {
            // Dateien loeschen
            for (i=0; i < j; i++) {
            
                sprintf(szFile, LOG_DELETE_FILENAME, m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder, szDeleteFileName[i]);
                //@LOG:
                printf("Datei wird gelöscht: %s\n", szFile);
                if (remove(szFile) == -1) {
                    // Konnte nicht geloescht werden!
                }
            }
        }
    }
*/
#endif
    
}

void CDFLog::logBegin(const char *Format, ...)
{
    if (!m_bEnable) {
    
        return;
    }

#ifdef _MSC_VER
    EnterCriticalSection( &m_mutexFile );
#else
    pthread_mutex_lock( &m_mutexFile );
#endif

    char szTime[9];
    char szFileName[260];
    char szLine[0x400];
    va_list ap;
    FILE *Logfp;

    // Uhrzeit des Logeintrages
    time_t t = time(NULL);
    m_clockBegin = clock();
    tm *ltm = localtime(&t);
    sprintf(szTime, "%02i:%02i:%02i", ltm->tm_hour, ltm->tm_min, ltm->tm_sec);

    // Dateiname der Logdatei
    sprintf(szFileName, LOG_FILENAME, m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder, 
        ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_channelID);
    
    if (strcmp(szFileName, m_szFileName) != 0) {
        // Dateiname hat sich geaendert.
        strcpy(m_szFileName, szFileName);
        // Dateiname auch fuer Tracefile anpassen
        sprintf(m_szFileNameTrace, LOG_FILENAME, m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder, 
            ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_channelID + 1000);
        // Setzen des Namens fuer Tracefile in abgeleiteter Klasse
        setTraceFileName(m_szFileNameTrace);
        logCheck(); // Eventuell die aelteste Logdatei loeschen.
    }

    sprintf(szLine, "%s\t", szTime);
#ifdef _MSC_VER
    // ProcessId und ThreadId einfuegen
    sprintf(&szLine[strlen(szLine)], "%u\t%u\t", static_cast<unsigned>( GetCurrentProcessId() ), static_cast<unsigned>( GetCurrentThreadId() ));
#endif
    va_start(ap, Format);
    vsprintf(&szLine[strlen(szLine)], Format, ap);
    va_end(ap);
    
    if ((Logfp = fopen(m_szFileName, "a")) == (FILE *)0) {

#ifdef _MSC_VER
        LeaveCriticalSection( &m_mutexFile );
#else
        pthread_mutex_unlock( &m_mutexFile );
#endif
        return;
    }

    fwrite(szLine, 1, strlen(szLine), Logfp);

    fclose(Logfp);

#ifdef _MSC_VER
    LeaveCriticalSection( &m_mutexFile );
#else
    pthread_mutex_unlock( &m_mutexFile );
#endif

}

void CDFLog::logAppend(const char *Format, ...)
{
    if (!m_bEnable) {

        return;
    }

#ifdef _MSC_VER
    EnterCriticalSection( &m_mutexFile );
#else
    pthread_mutex_lock( &m_mutexFile );
#endif

    va_list ap;
    FILE *Logfp;

    if ((Logfp = fopen(m_szFileName, "a")) == (FILE *)0) {

#ifdef _MSC_VER
        LeaveCriticalSection( &m_mutexFile );
#else
        pthread_mutex_unlock( &m_mutexFile );
#endif
        return;
    }

    va_start(ap, Format);
    vfprintf(Logfp, Format, ap);
    va_end(ap);

    fclose(Logfp);

#ifdef _MSC_VER
    LeaveCriticalSection( &m_mutexFile );
#else
    pthread_mutex_unlock( &m_mutexFile );
#endif

}

void CDFLog::logEnd(const char *Format, ...)
{
    if (!m_bEnable) {
        
        return;
    }

#ifdef _MSC_VER
    EnterCriticalSection( &m_mutexFile );
#else
    pthread_mutex_lock( &m_mutexFile );
#endif

    va_list ap;
    FILE *Logfp;

    if ((Logfp = fopen(m_szFileName, "a")) == (FILE *)0) {

#ifdef _MSC_VER
        LeaveCriticalSection( &m_mutexFile );
#else
        pthread_mutex_unlock( &m_mutexFile );
#endif
        return;
    }

    va_start(ap, Format);
    vfprintf(Logfp, Format, ap);
    va_end(ap);

    // Ausfuehrungszeit angeben?
    if (m_bElapse) {
        // Ausfuehrungszeit berechnen
        double clockDiff;
        clockDiff = (double)(clock() - m_clockBegin) / CLOCKS_PER_SEC;
#ifdef _MSC_VER
        fprintf(Logfp, "\t[%2.1f]\n", clockDiff);
#else
        fprintf(Logfp, "\t[%2.1f]\r\n", clockDiff);
#endif
    } else {
#ifdef _MSC_VER
        fprintf(Logfp, "\n");
#else
        fprintf(Logfp, "\r\n");
#endif
    }

    fclose(Logfp);

#ifdef _MSC_VER
    LeaveCriticalSection( &m_mutexFile );
#else
    pthread_mutex_unlock( &m_mutexFile );
#endif
}

#ifdef _MSC_VER
void CDFLog::MakeFilePath(char *path_buffer, char *szFileName, char *szFileExt)
{
    char drive[_MAX_DRIVE];
    char dir[_MAX_DIR];
    char fname[_MAX_FNAME];
    char ext[_MAX_EXT];
    
    // Speicher initialisieren
    memset(drive, 0, sizeof(drive));
    memset(dir, 0, sizeof(dir));
    memset(fname, 0, sizeof(fname));
    memset(ext, 0, sizeof(ext));
    
    _splitpath(path_buffer, drive, dir, fname, ext);
    // Wenn keine Extension, dann gehoert eventueller Dateiname noch zum Pfad
    if (!strlen(ext)) {
        
        if (strlen(fname)) {
            
            strcat(path_buffer, "\\");
            // Neu auswerten
            _splitpath(path_buffer, drive, dir, fname, ext);
        }
    }
    
    // Speicher loeschen
    memset(path_buffer, 0, _MAX_PATH);
    
    _makepath(path_buffer, drive, dir, (szFileName) ? szFileName : fname, (szFileExt) ? szFileExt : ext);
}
#endif


